"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import importlib
import logging
import os
import traceback
from datetime import datetime
from typing import Any, Dict, List, MutableSet, Optional, Tuple, Union

import cfnlint.helpers
import cfnlint.rules.custom
from cfnlint.decode.node import TemplateAttributeError
from cfnlint.exceptions import DuplicateRuleError
from cfnlint.template import Template

LOGGER = logging.getLogger(__name__)


def matching(match_type: Any):
    """Does Logging for match functions"""

    def decorator(match_function):
        """The Actual Decorator"""

        def wrapper(self, filename: str, cfn: Template, *args, **kwargs):
            """Wrapper"""
            matches = []

            if not getattr(self, match_type):
                return []

            if match_type == "match_resource_properties":
                if args[1] not in self.resource_property_types:
                    return []
            elif match_type == "match_resource_sub_properties":
                if args[1] not in self.resource_sub_property_types:
                    return []

            start = datetime.now()
            LOGGER.debug("Starting match function for rule %s at %s", self.id, start)
            # pylint: disable=E1102
            results = match_function(self, filename, cfn, *args, **kwargs)
            LOGGER.debug(
                "Complete match function for rule %s at %s.  Ran in %s",
                self.id,
                datetime.now(),
                datetime.now() - start,
            )
            LOGGER.debug("Results from rule %s are %s: ", self.id, results)

            if results:
                for result in results:
                    error_rule = self
                    if hasattr(result, "rule"):
                        error_rule = result.rule
                    linenumbers: Union[Tuple[int, int, int, int], None] = None
                    if hasattr(result, "location"):
                        linenumbers = result.location
                    else:
                        linenumbers = cfn.get_location_yaml(cfn.template, result.path)
                    if linenumbers:
                        matches.append(
                            Match(
                                linenumbers[0] + 1,
                                linenumbers[1] + 1,
                                linenumbers[2] + 1,
                                linenumbers[3] + 1,
                                filename,
                                error_rule,
                                result.message,
                                result,
                            )
                        )
                    else:
                        matches.append(
                            Match(
                                1, 1, 1, 1, filename, error_rule, result.message, result
                            )
                        )

            return matches

        return wrapper

    return decorator


class CloudFormationLintRule:
    """CloudFormation linter rules"""

    id: str = ""
    shortdesc: str = ""
    description: str = ""
    source_url: str = ""
    tags: List[str] = []
    experimental: bool = False
    child_rules: Dict[str, Any] = {}

    logger = logging.getLogger(__name__)

    def __init__(self):
        self.resource_property_types = []
        self.resource_sub_property_types = []
        self.config = {}  # `-X E3012:strict=false`... Show more
        self.config_definition = {}

    def __repr__(self):
        return f"{self.id}: {self.shortdesc}"

    @property
    def severity(self):
        """Severity level"""
        levels = {
            "I": "informational",
            "E": "error",
            "W": "warning",
        }
        return levels.get(self.id[0].upper(), "unknown")

    def verbose(self):
        """Verbose output"""
        return f"{self.id}: {self.shortdesc}\n{self.description}"

    def initialize(self, cfn):
        """Initialize the rule"""

    def is_enabled(
        self,
        include_experimental=False,
        ignore_rules=None,
        include_rules=None,
        mandatory_rules=None,
    ):
        """Is the rule enabled based on the configuration"""
        ignore_rules = ignore_rules or []
        include_rules = include_rules or []
        mandatory_rules = mandatory_rules or []

        # Evaluate experimental rules
        if self.experimental and not include_experimental:
            return False

        # Evaluate includes first:
        include_filter = False
        for include_rule in include_rules:
            if self.id.startswith(include_rule):
                include_filter = True
        if not include_filter:
            return False

        # Enable mandatory rules without checking for if they are ignored
        for mandatory_rule in mandatory_rules:
            if self.id.startswith(mandatory_rule):
                return True

        # Allowing ignoring of rules based on prefix to ignore checks
        for ignore_rule in ignore_rules:
            if self.id.startswith(ignore_rule) and ignore_rule:
                return False

        return True

    def configure(self, configs=None):
        """Set the configuration"""

        # set defaults
        if isinstance(self.config_definition, dict):
            for config_name, config_values in self.config_definition.items():
                self.config[config_name] = config_values["default"]

        if isinstance(configs, dict):
            for key, value in configs.items():
                if key in self.config_definition:
                    if self.config_definition[key]["type"] == "boolean":
                        self.config[key] = cfnlint.helpers.bool_compare(value, True)
                    elif self.config_definition[key]["type"] == "string":
                        self.config[key] = str(value)
                    elif self.config_definition[key]["type"] == "integer":
                        self.config[key] = int(value)
                    elif self.config_definition[key]["type"] == "list":
                        self.config[key] = []
                        for l_value in value:
                            if self.config_definition[key]["itemtype"] == "boolean":
                                self.config[key].append(
                                    cfnlint.helpers.bool_compare(l_value, True)
                                )
                            elif self.config_definition[key]["itemtype"] == "string":
                                self.config[key].append(str(l_value))
                            elif self.config_definition[key]["itemtype"] == "integer":
                                self.config[key].append(int(l_value))

    match = None
    match_resource_properties = None
    match_resource_sub_properties = None

    @matching("match")
    # pylint: disable=W0613
    def matchall(self, filename, cfn):
        """Match the entire file"""
        return self.match(cfn)  # pylint: disable=E1102

    @matching("match_resource_properties")
    # pylint: disable=W0613
    def matchall_resource_properties(
        self, filename, cfn, resource_properties, property_type, path
    ):
        """Check for resource properties type"""
        return self.match_resource_properties(  # pylint: disable=E1102
            resource_properties, property_type, path, cfn
        )

    @matching("match_resource_sub_properties")
    # pylint: disable=W0613
    def matchall_resource_sub_properties(
        self, filename, cfn, resource_properties, property_type, path
    ):
        """Check for resource properties type"""
        return self.match_resource_sub_properties(  # pylint: disable=E1102
            resource_properties, property_type, path, cfn
        )


# pylint: disable=too-many-instance-attributes
class RulesCollection:
    """Collection of rules"""

    def __init__(
        self,
        ignore_rules: Union[List[str], None] = None,
        include_rules: Union[List[str], None] = None,
        configure_rules: Any = None,
        include_experimental: bool = False,
        mandatory_rules: Union[List[str], None] = None,
    ):
        self.rules: Dict[str, CloudFormationLintRule] = {}
        self.all_rules: Dict[str, CloudFormationLintRule] = {}
        self.used_rules: MutableSet[str] = set()

        self.configure(
            ignore_rules=ignore_rules,
            include_rules=include_rules,
            configure_rules=configure_rules,
            include_experimental=include_experimental,
            mandatory_rules=mandatory_rules,
        )

    def configure(
        self,
        ignore_rules: Union[List[str], None] = None,
        include_rules: Union[List[str], None] = None,
        configure_rules: Any = None,
        include_experimental: bool = False,
        mandatory_rules: Union[List[str], None] = None,
    ):
        self.rules = {}
        # Whether "experimental" rules should be added
        self.include_experimental = include_experimental

        # Make Ignore Rules not required
        self.ignore_rules = ignore_rules or []
        self.include_rules = include_rules or []
        self.mandatory_rules = mandatory_rules or []
        self.configure_rules = configure_rules or {}
        # by default include 'W' and 'E'
        # 'I' has to be included manually for backwards compabitility
        # Have to add W, E here because integrations don't use config
        for default_rule in ["W", "E"]:
            if default_rule not in self.include_rules:
                self.include_rules.extend([default_rule])

        for rule in self.all_rules.values():
            self.__register(rule)

    def __register(self, rule: CloudFormationLintRule):
        """Register and configure the rule"""
        if self.is_rule_enabled(rule):
            self.used_rules.add(rule.id)
            self.rules[rule.id] = rule
            rule.configure(self.configure_rules.get(rule.id, None))

    def register(self, rule: CloudFormationLintRule):
        """Register rules"""
        # Some rules are inheritited to limit code re-use.
        # These rules have no rule ID so we filter this out
        if rule.id != "":
            if rule.id in self.all_rules:
                raise DuplicateRuleError(rule_id=rule.id)
            self.all_rules[rule.id] = rule
            self.__register(rule)

    def __iter__(self):
        return iter(self.rules.values())

    def __len__(self):
        return len(self.rules.keys())

    def extend(self, more):
        """Extend rules"""
        for rule in more:
            self.register(rule)

    def __repr__(self):
        return "\n".join([self.rules[id].verbose() for id in sorted(self.rules)])

    def is_rule_enabled(self, rule: CloudFormationLintRule):
        """Checks if an individual rule is valid"""
        return rule.is_enabled(
            self.include_experimental,
            self.ignore_rules,
            self.include_rules,
            self.mandatory_rules,
        )

    # pylint: disable=inconsistent-return-statements
    def run_check(self, check, filename, rule_id, *args):
        """Run a check"""
        try:
            return check(*args)
        except TemplateAttributeError as err:
            LOGGER.debug(str(err))
            return []
        except Exception as err:  # pylint: disable=W0703
            if self.is_rule_enabled(RuleError()):
                # In debug mode, print the error include complete stack trace
                if LOGGER.getEffectiveLevel() == logging.DEBUG:
                    error_message = traceback.format_exc()
                else:
                    error_message = str(err)
                message = "Unknown exception while processing rule {}: {}"
                return [
                    Match(
                        1,
                        1,
                        1,
                        1,
                        filename,
                        RuleError(),
                        message.format(rule_id, error_message),
                    )
                ]

    def resource_property(
        self, filename, cfn, path, properties, resource_type, property_type
    ):
        """Run loops in resource checks for embedded properties"""
        matches = []
        property_spec = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("PropertyTypes")
        if property_type == "Tag":
            property_spec_name = "Tag"
        else:
            property_spec_name = f"{resource_type}.{property_type}"

        if property_spec_name in property_spec:
            for rule in self.rules.values():
                if isinstance(properties, dict):
                    if len(properties) == 1:
                        for k, _ in properties.items():
                            if k != "Fn::If":
                                matches.extend(
                                    self.run_check(
                                        rule.matchall_resource_sub_properties,
                                        filename,
                                        rule.id,
                                        filename,
                                        cfn,
                                        properties,
                                        property_spec_name,
                                        path,
                                    )
                                )
                    else:
                        matches.extend(
                            self.run_check(
                                rule.matchall_resource_sub_properties,
                                filename,
                                rule.id,
                                filename,
                                cfn,
                                properties,
                                property_spec_name,
                                path,
                            )
                        )
                else:
                    matches.extend(
                        self.run_check(
                            rule.matchall_resource_sub_properties,
                            filename,
                            rule.id,
                            filename,
                            cfn,
                            properties,
                            property_spec_name,
                            path,
                        )
                    )

            resource_spec_properties = property_spec.get(property_spec_name, {}).get(
                "Properties"
            )
            if not resource_spec_properties:
                if property_spec.get(property_spec_name, {}).get("Type") == "List":
                    if isinstance(properties, list):
                        property_type = property_spec.get(property_spec_name, {}).get(
                            "ItemType"
                        )
                        for index, item in enumerate(properties):
                            matches.extend(
                                self.resource_property(
                                    filename,
                                    cfn,
                                    path[:] + [index],
                                    item,
                                    resource_type,
                                    property_type,
                                )
                            )
                return matches
            if isinstance(properties, dict):
                for resource_property, resource_property_value in properties.items():
                    property_path = path[:] + [resource_property]
                    resource_spec_property = resource_spec_properties.get(
                        resource_property, {}
                    )
                    if resource_property not in resource_spec_properties:
                        if resource_property == "Fn::If":
                            if isinstance(resource_property_value, list):
                                if len(resource_property_value) == 3:
                                    for index, c_value in enumerate(
                                        resource_property_value[1:]
                                    ):
                                        if isinstance(c_value, list):
                                            for s_i, c_l_value in enumerate(c_value):
                                                matches.extend(
                                                    self.resource_property(
                                                        filename,
                                                        cfn,
                                                        property_path[:]
                                                        + [index + 1]
                                                        + [s_i],
                                                        c_l_value,
                                                        resource_type,
                                                        property_type,
                                                    )
                                                )
                                        else:
                                            matches.extend(
                                                self.resource_property(
                                                    filename,
                                                    cfn,
                                                    property_path[:] + [index + 1],
                                                    c_value,
                                                    resource_type,
                                                    property_type,
                                                )
                                            )
                        continue
                    if resource_spec_property.get(
                        "Type"
                    ) == "List" and not resource_spec_properties.get(
                        "PrimitiveItemType"
                    ):
                        if isinstance(resource_property_value, (list)):
                            for index, value in enumerate(resource_property_value):
                                matches.extend(
                                    self.resource_property(
                                        filename,
                                        cfn,
                                        property_path[:] + [index],
                                        value,
                                        resource_type,
                                        resource_spec_property.get("ItemType"),
                                    )
                                )
                    elif resource_spec_property.get("Type"):
                        if isinstance(resource_property_value, (dict)):
                            matches.extend(
                                self.resource_property(
                                    filename,
                                    cfn,
                                    property_path,
                                    resource_property_value,
                                    resource_type,
                                    resource_spec_property.get("Type"),
                                )
                            )

        return matches

    def run_resource(self, filename, cfn, resource_type, resource_properties, path):
        """Run loops in resource checks for embedded properties"""
        matches = []
        resource_spec = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("ResourceTypes")
        if resource_properties is not None and resource_type in resource_spec:
            resource_spec_properties = resource_spec.get(resource_type, {}).get(
                "Properties"
            )
            items_safe = resource_properties.items_safe(path, type_t=dict)
            for resource_properties_safe, path_safe in items_safe:
                for (
                    resource_property,
                    resource_property_value,
                ) in resource_properties_safe.items():
                    resource_spec_property = resource_spec_properties.get(
                        resource_property, {}
                    )
                    if resource_spec_property.get(
                        "Type"
                    ) == "List" and not resource_spec_properties.get(
                        "PrimitiveItemType"
                    ):
                        if isinstance(resource_property_value, (list)):
                            for index, value in enumerate(resource_property_value):
                                matches.extend(
                                    self.resource_property(
                                        filename,
                                        cfn,
                                        path_safe[:] + [resource_property, index],
                                        value,
                                        resource_type,
                                        resource_spec_property.get("ItemType"),
                                    )
                                )
                    elif resource_spec_property.get("Type"):
                        if isinstance(resource_property_value, (dict)):
                            matches.extend(
                                self.resource_property(
                                    filename,
                                    cfn,
                                    path_safe[:] + [resource_property],
                                    resource_property_value,
                                    resource_type,
                                    resource_spec_property.get("Type"),
                                )
                            )

        return matches

    def run(self, filename: Optional[str], cfn: Template):
        """Run rules"""
        matches = []
        for rule in self.rules.values():
            rule.initialize(cfn)

        for rule in self.rules.values():
            for key in rule.child_rules.keys():
                rule.child_rules[key] = self.rules.get(key)

        for rule in self.rules.values():
            matches.extend(
                self.run_check(rule.matchall, filename, rule.id, filename, cfn)
            )

        for resource_name, resource_attributes in cfn.get_resources().items():
            resource_type = resource_attributes.get("Type")
            resource_properties = resource_attributes.get("Properties")
            if isinstance(resource_type, str) and isinstance(resource_properties, dict):
                path = ["Resources", resource_name, "Properties"]
                for rule in self.rules.values():
                    matches.extend(
                        self.run_check(
                            rule.matchall_resource_properties,
                            filename,
                            rule.id,
                            filename,
                            cfn,
                            resource_properties,
                            resource_type,
                            path,
                        )
                    )

                matches.extend(
                    self.run_resource(
                        filename, cfn, resource_type, resource_properties, path
                    )
                )

        return matches

    def create_from_module(self, modpath):
        """Create rules from a module import path"""
        mod = importlib.import_module(modpath)
        self.extend(cfnlint.helpers.create_rules(mod))

    def create_from_directory(self, rulesdir):
        """Create rules from directory"""
        result = []
        if rulesdir != "":
            result = cfnlint.helpers.load_plugins(os.path.expanduser(rulesdir))
        self.extend(result)

    def create_from_custom_rules_file(self, custom_rules_file):
        """Create rules from custom rules file"""
        custom_rules = []
        if custom_rules_file:
            with open(custom_rules_file, encoding="utf-8") as customRules:
                line_number = 1
                for line in customRules:
                    LOGGER.debug("Processing Custom Rule Line %d", line_number)
                    custom_rule = cfnlint.rules.custom.make_rule(line, line_number)
                    if custom_rule:
                        custom_rules.append(custom_rule)
                    line_number += 1

        self.extend(custom_rules)


class RuleMatch:
    """Rules Error"""

    def __init__(self, path, message, **kwargs):
        """Init"""
        self.path = path
        self.path_string = "/".join(map(str, path))
        self.message = message
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, item):
        """Override unique"""
        return (self.path, self.message) == (item.path, item.message)

    def __hash__(self):
        """Hash for comparisons"""
        return hash((self.path, self.message))


class Match:  # pylint: disable=R0902
    """Match Classes"""

    def __init__(
        self,
        linenumber,
        columnnumber,
        linenumberend,
        columnnumberend,
        filename,
        rule,
        message=None,
        rulematch_obj=None,
    ):
        """Init"""
        self.linenumber = linenumber
        """Starting line number of the region this match spans"""
        self.columnnumber = columnnumber
        """Starting line number of the region this match spans"""
        self.linenumberend = linenumberend
        """Ending line number of the region this match spans"""
        self.columnnumberend = columnnumberend
        """Ending column number of the region this match spans"""
        self.filename = filename
        """Name of the filename associated with this match, or None if there is no such file"""
        self.rule = rule
        """The rule of this match"""
        self.message = message  # or rule.shortdesc
        """The message of this match"""
        if rulematch_obj:
            for k, v in vars(rulematch_obj).items():
                if not hasattr(self, k):
                    setattr(self, k, v)

    def __repr__(self):
        """Represent"""
        file_str = self.filename + ":" if self.filename else ""
        return f"[{self.rule}] ({self.message}) matched {file_str}{self.linenumber}"

    def __eq__(self, item):
        """Override equal to compare matches"""
        return (self.linenumber, self.columnnumber, self.rule.id, self.message) == (
            item.linenumber,
            item.columnnumber,
            item.rule.id,
            item.message,
        )


class ParseError(CloudFormationLintRule):
    """Parse Lint Rule"""

    id = "E0000"
    shortdesc = "Parsing error found when parsing the template"
    description = "Checks for JSON/YAML formatting errors in your template"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["base"]


class TransformError(CloudFormationLintRule):
    """Transform Lint Rule"""

    id = "E0001"
    shortdesc = "Error found when transforming the template"
    description = "Errors found when performing transformation on the template"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["base", "transform"]


class RuleError(CloudFormationLintRule):
    """Rule processing Error"""

    id = "E0002"
    shortdesc = "Error processing rule on the template"
    description = "Errors found when processing a rule on the template"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["base", "rule"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.helpers import REGEX_ALPHANUMERIC
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class KeyName(CloudFormationLintRule):
    """Check if Mapping Keys are type string"""

    id = "E7003"
    shortdesc = "Mapping keys are strings and alphanumeric"
    description = (
        "Check if Mappings keys are properly typed as strings and alphanumeric"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html"
    tags = ["mappings"]

    def check_attribute(self, key, path):
        """Check the key name for string and alphanumeric"""
        matches = []
        if not isinstance(key, str):
            message = "Mapping attribute ({0}) has to be a string."
            matches.append(RuleMatch(path[:], message.format(key)))
        elif not re.match(REGEX_ALPHANUMERIC, key):
            message = (
                "Mapping attribute ({0}) has invalid name. Name has to be alphanumeric."
            )
            matches.append(RuleMatch(path[:], message.format(key)))

        return matches

    def check_key(self, key, path):
        """Check the key name for string and alphanumeric"""
        matches = []
        if not isinstance(key, str):
            message = "Mapping key ({0}) has to be a string."
            matches.append(RuleMatch(path[:], message.format(key)))
        elif not re.match("^[a-zA-Z0-9.-]{1,255}$", key) and key != "Fn::Transform":
            message = "Mapping key ({0}) has invalid name. Name has to be alphanumeric, '-' or '.'"
            matches.append(RuleMatch(path[:], message.format(key)))

        return matches

    def match(self, cfn):
        matches = []

        mappings = cfn.template.get("Mappings", {})
        for mapping_name, mapping_value in mappings.items():
            if isinstance(mapping_value, dict):
                for key_name, key_value in mapping_value.items():
                    matches.extend(
                        self.check_key(key_name, ["Mappings", mapping_name, key_name])
                    )
                    if isinstance(key_value, dict):
                        for sub_key_name, _ in key_value.items():
                            matches.extend(
                                self.check_attribute(
                                    sub_key_name,
                                    ["Mappings", mapping_name, key_name, sub_key_name],
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import itertools
import json
import operator
import re
import sys
from typing import List

import sarif_om as sarif
from jschema_to_python.to_json import to_json
from junit_xml import TestCase, TestSuite, to_xml_report_string

from cfnlint.rules import Match, ParseError, RuleError, RulesCollection, TransformError
from cfnlint.version import __version__

Matches = List[Match]


class color:
    error = "\033[31m"
    warning = "\033[33m"
    informational = "\033[34m"
    unknown = "\033[37m"
    green = "\033[32m"
    reset = "\033[0m"
    bold_reset = "\033[1:0m"
    underline_reset = "\033[4m"


def colored(s, c):
    """Takes in string s and outputs it with color"""
    if sys.stdout.isatty():
        return f"{c}{s}{color.reset}"

    return s


class BaseFormatter:
    """Base Formatter class"""

    def _format(self, match):
        """Format the specific match"""

    def print_matches(self, matches, rules=None, filenames=None):
        """Output all the matches"""
        # Unused argument http://pylint-messages.wikidot.com/messages:w0613
        del rules
        del filenames

        if not matches:
            return None

        # Output each match on a separate line by default
        output = []
        for match in matches:
            output.append(self._format(match))

        return "\n".join(output)


class Formatter(BaseFormatter):
    """Generic Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0} {1}\n{2}:{3}:{4}\n"
        return formatstr.format(
            match.rule.id,
            match.message,
            match.filename,
            match.linenumber,
            match.columnnumber,
        )


class JUnitFormatter(BaseFormatter):
    """JUnit-style Reports"""

    def _failure_format(self, match):
        """Format output of a failure"""
        formatstr = "{0} at {1}:{2}:{3}"
        return formatstr.format(
            match.message, match.filename, match.linenumber, match.columnnumber
        )

    def print_matches(self, matches, rules=None, filenames=None):
        """Output all the matches"""

        if not rules:
            return None

        if rules is not None:
            # These "base" rules are not passed into formatters
            rules.extend([ParseError(), TransformError(), RuleError()])

        test_cases = []
        for rule in rules.all_rules.values():
            if not rule.id in rules.used_rules:
                if not rule.id:
                    continue
                test_case = TestCase(name=f"{rule.id} {rule.shortdesc}")

                if rule.experimental:
                    test_case.add_skipped_info(
                        message="Experimental rule - not enabled"
                    )
                else:
                    test_case.add_skipped_info(message="Ignored rule")
                test_cases.append(test_case)
            else:
                test_case = TestCase(
                    name=f"{rule.id} {rule.shortdesc}",
                    allow_multiple_subelements=True,
                    url=rule.source_url,
                )
                for match in matches:
                    if match.rule.id == rule.id:
                        test_case.add_failure_info(
                            message=self._failure_format(match),
                            failure_type=match.message,
                        )
                test_cases.append(test_case)

        test_suite = TestSuite("CloudFormation Lint", test_cases)

        return to_xml_report_string([test_suite], prettyprint=True)


class JsonFormatter(BaseFormatter):
    """Json Formatter"""

    class CustomEncoder(json.JSONEncoder):
        """Custom Encoding for the Match Object"""

        # pylint: disable=E0202

        def default(self, o):
            if isinstance(o, Match):
                return {
                    "Rule": {
                        "Id": o.rule.id,
                        "Description": o.rule.description,
                        "ShortDescription": o.rule.shortdesc,
                        "Source": o.rule.source_url,
                    },
                    "Location": {
                        "Start": {
                            "ColumnNumber": o.columnnumber,
                            "LineNumber": o.linenumber,
                        },
                        "End": {
                            "ColumnNumber": o.columnnumberend,
                            "LineNumber": o.linenumberend,
                        },
                        "Path": getattr(o, "path", None),
                    },
                    "Level": o.rule.severity.capitalize(),
                    "Message": o.message,
                    "Filename": o.filename,
                }
            return {f"__{o.__class__.__name__}__": o.__dict__}

    def print_matches(self, matches, rules=None, filenames=None):
        # JSON formatter outputs a single JSON object
        # Unused argument http://pylint-messages.wikidot.com/messages:w0613
        del rules

        return json.dumps(
            matches,
            indent=4,
            cls=self.CustomEncoder,
            sort_keys=True,
            separators=(",", ": "),
        )


class QuietFormatter(BaseFormatter):
    """Quiet Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0} {1}:{2}"
        return formatstr.format(match.rule, match.filename, match.linenumber)


class ParseableFormatter(BaseFormatter):
    """Parseable Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0}:{1}:{2}:{3}:{4}:{5}:{6}"
        return formatstr.format(
            match.filename,
            match.linenumber,
            match.columnnumber,
            match.linenumberend,
            match.columnnumberend,
            match.rule.id,
            re.sub(r"(\r*\n)+", " ", match.message),
        )


class PrettyFormatter(BaseFormatter):
    """Generic Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0}{1}{2}"
        pos = f"{match.linenumber}:{match.columnnumber}:"
        return formatstr.format(
            colored(f"{pos:20}", color.reset),
            colored(f"{match.rule.id:10}", getattr(color, match.rule.severity.lower())),
            match.message,
        )

    def print_matches(self, matches, rules=None, filenames=None):
        results = self._format_matches(matches)

        results.append(
            f"Cfn-lint scanned {colored(len(filenames), color.bold_reset)} templates against "
            f"{colored(len(rules.used_rules), color.bold_reset)} rules and found "
            f'{colored(len([i for i in matches if i.rule.severity.lower() == "error"]), color.error)} '
            f'errors, {colored(len([i for i in matches if i.rule.severity.lower() == "warning"]), color.warning)} '
            f"warnings, and "
            f'{colored(len([i for i in matches if i.rule.severity.lower() == "informational"]), color.informational)} '
            f"informational violations"
        )
        return "\n".join(results)

    def _format_matches(self, matches):
        """Output all the matches"""
        output = []

        # This better be sorted
        for filename, file_matches in itertools.groupby(
            matches, key=operator.attrgetter("filename")
        ):
            levels = {"error": [], "warning": [], "informational": [], "unknown": []}

            output.append(colored(filename, color.underline_reset))
            for match in file_matches:
                level = match.rule.severity.lower()
                if level not in ["error", "warning", "informational"]:
                    level = "unknown"
                levels[level].append(match)
            for _, all_matches in levels.items():
                for match in all_matches:
                    output.extend([self._format(match)])

            output.append("")  # Newline after each group

        return output


class SARIFFormatter(BaseFormatter):
    """
    SARIF formatter

    This formatter outputs results according to the Static Analysis Results
    Interchange Format (SARIF) Version 2.1.0.

    https://docs.oasis-open.org/sarif/sarif/v2.1.0/csprd01/sarif-v2.1.0-csprd01.html
    """

    schema = "https://docs.oasis-open.org/sarif/sarif/v2.1.0/cos02/schemas/sarif-schema-2.1.0.json"
    version = "2.1.0"

    # The spec defines error, note, warning, and none, see section 3.27.10.
    levelMap = {
        "error": "error",
        "informational": "note",
        "warning": "warning",
    }

    uri_base_id = "EXECUTIONROOT"

    def _to_sarif_level(self, severity):
        return self.levelMap.get(severity, "none")

    def print_matches(self, matches, rules=None, filenames=None):
        """Output all the matches"""

        if not rules:
            rules = RulesCollection()

        # These "base" rules are not passed into formatters
        rules.extend([ParseError(), TransformError(), RuleError()])

        results = []
        for match in matches:
            results.append(
                sarif.Result(
                    rule_id=match.rule.id,
                    message=sarif.Message(text=match.message),
                    level=self._to_sarif_level(match.rule.severity),
                    locations=[
                        sarif.Location(
                            physical_location=sarif.PhysicalLocation(
                                artifact_location=sarif.ArtifactLocation(
                                    uri=match.filename,
                                    uri_base_id=self.uri_base_id,
                                ),
                                region=sarif.Region(
                                    start_column=match.columnnumber,
                                    start_line=match.linenumber,
                                    end_column=match.columnnumberend,
                                    end_line=match.linenumberend,
                                ),
                            )
                        )
                    ],
                )
            )

        # Output only the rules that have matches
        matched_rules = set(r.rule_id for r in results)
        rules_map = {r.id: r for r in list(rules)}

        rules = [
            sarif.ReportingDescriptor(
                id=rule_id,
                short_description=sarif.MultiformatMessageString(
                    text=rules_map[rule_id].shortdesc
                ),
                full_description=sarif.MultiformatMessageString(
                    text=rules_map[rule_id].description
                ),
                help_uri=rules_map[rule_id].source_url
                if rules_map[rule_id]
                else "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/rules.md",
            )
            for rule_id in matched_rules
        ]

        run = sarif.Run(
            tool=sarif.Tool(
                driver=sarif.ToolComponent(
                    name="cfn-lint",
                    short_description=sarif.MultiformatMessageString(
                        text=(
                            "Validates AWS CloudFormation templates against"
                            " the resource specification and additional"
                            " checks."
                        )
                    ),
                    information_uri="https://github.com/aws-cloudformation/cfn-lint",
                    rules=rules,
                    version=__version__,
                ),
            ),
            original_uri_base_ids={
                self.uri_base_id: sarif.ArtifactLocation(
                    description=sarif.MultiformatMessageString(
                        "The directory in which cfn-lint was run."
                    )
                )
            },
            results=results,
        )

        log = sarif.SarifLog(version=self.version, schema_uri=self.schema, runs=[run])

        # IMPORTANT: 'warning' is the default level in SARIF and will be
        # stripped by serialization.
        return to_json(log)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ImageId(CloudFormationLintRule):
    id = "W2506"
    shortdesc = "Check if ImageId Parameters have the correct type"
    description = (
        "See if there are any refs for ImageId to a parameter "
        + "of inappropriate type. Appropriate Types are "
        + "[AWS::EC2::Image::Id, AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>]"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["parameters", "ec2", "imageid"]

    def match(self, cfn):
        """Check CloudFormation ImageId Parameters"""

        matches = []

        # Build the list of refs
        imageidtrees = cfn.search_deep_keys("ImageId")
        valid_refs = cfn.get_valid_refs()
        allowed_types = [
            "AWS::EC2::Image::Id",
            "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
        ]
        # Filter only resoureces
        imageidtrees = [x for x in imageidtrees if x[0] == "Resources"]
        for imageidtree in imageidtrees:
            imageidobj = imageidtree[-1]
            if isinstance(imageidobj, dict):
                if len(imageidobj) == 1:
                    for key, paramname in imageidobj.items():
                        if key == "Ref":
                            if paramname in valid_refs:
                                if valid_refs[paramname]["From"] == "Parameters":
                                    if (
                                        valid_refs[paramname]["Type"]
                                        not in allowed_types
                                    ):
                                        message = (
                                            f"Parameter {paramname} should be of type "
                                            f'[{", ".join(map(str, allowed_types))}]'
                                        )
                                        tree = ["Parameters", paramname]
                                        matches.append(RuleMatch(tree, message))
                else:
                    message = f'Inappropriate map found for ImageId on {"/".join(map(str, imageidtree[:-1]))}'
                    matches.append(RuleMatch(imageidtree[:-1], message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DependsOn(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = "E3005"
    shortdesc = "Check DependsOn values for Resources"
    description = "Check that the DependsOn values are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-dependson.html"
    tags = ["resources", "dependson"]

    def check_value(self, key, path, resources, cfn):
        """Check resource names for DependsOn"""
        matches = []

        if not isinstance(key, (str)):
            message = "DependsOn values should be of string at {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
            return matches
        if key not in resources:
            message = "DependsOn should reference other resources at {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))
        else:
            for scenario in cfn.is_resource_available(path, key):
                if scenario:
                    scenario_text = " and ".join(
                        [
                            f'when condition "{k}" is {scenario[k]}'
                            for k in sorted(scenario)
                        ]
                    )
                    message = "DependsOn {0} may not exist when condition {1} at {2}"
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(
                                key, scenario_text, "/".join(map(str, path))
                            ),
                        )
                    )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.get_resources()

        for resource_name, resource_values in resources.items():
            depends_ons = resource_values.get("DependsOn")
            if depends_ons:
                path = ["Resources", resource_name, "DependsOn"]
                self.logger.debug(
                    "Validating DependsOn for %s base configuration", resource_name
                )
                if isinstance(depends_ons, list):
                    for index, depends_on in enumerate(depends_ons):
                        matches.extend(
                            self.check_value(
                                depends_on, path[:] + [index], resources, cfn
                            )
                        )
                else:
                    matches.extend(self.check_value(depends_ons, path, resources, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.decode.decode import (
    create_match_file_error,
    create_match_json_parser_error,
    create_match_yaml_parser_error,
    decode,
    decode_str,
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.decode.node import sub_node
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SubUnneeded(CloudFormationLintRule):
    """Check if Sub is using a variable"""

    id = "W1020"
    shortdesc = "Sub isn't needed if it doesn't have a variable defined"
    description = "Checks sub strings to see if a variable is defined."
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def match(self, cfn):
        matches = []

        sub_objs = cfn.search_deep_class(sub_node)

        for sub_obj_tuple in sub_objs:
            path, sub_obj = sub_obj_tuple
            result_path = path[:] + ["Fn::Sub"]
            if sub_obj.is_valid():
                if not sub_obj.get_string_vars():
                    message = (
                        "Fn::Sub isn't needed because there are no variables at {0}"
                    )
                    matches.append(
                        RuleMatch(
                            result_path, message.format("/".join(map(str, result_path)))
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Split(CloudFormationLintRule):
    """Check if Split values are correct"""

    id = "E1018"
    shortdesc = "Split validation of parameters"
    description = "Making sure the split function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-split.html"
    tags = ["functions", "split"]

    def match(self, cfn):
        matches = []

        split_objs = cfn.search_deep_keys("Fn::Split")

        supported_functions = [
            "Fn::Base64",
            "Fn::FindInMap",
            "Fn::GetAZs",
            "Fn::GetAtt",
            "Fn::If",
            "Fn::ImportValue",
            "Fn::Join",
            "Fn::Select",
            "Fn::Sub",
            "Ref",
            "Fn::ToJsonString",
        ]

        for split_obj in split_objs:
            split_value_obj = split_obj[-1]
            tree = split_obj[:-1]
            if isinstance(split_value_obj, list):
                if len(split_value_obj) == 2:
                    split_delimiter = split_value_obj[0]
                    split_string = split_value_obj[1]
                    if not isinstance(split_delimiter, str):
                        message = "Split delimiter has to be of type string for {0}"
                        matches.append(
                            RuleMatch(
                                tree + [0], message.format("/".join(map(str, tree)))
                            )
                        )
                    if isinstance(split_string, dict):
                        if len(split_string) == 1:
                            for key, _ in split_string.items():
                                if key not in supported_functions:
                                    message = "Fn::Split doesn't support the function {0} at {1}"
                                    matches.append(
                                        RuleMatch(
                                            tree + [key],
                                            message.format(
                                                key, "/".join(map(str, tree))
                                            ),
                                        )
                                    )
                        else:
                            message = (
                                "Split list of singular function or string for {0}"
                            )
                            matches.append(
                                RuleMatch(
                                    tree, message.format("/".join(map(str, tree)))
                                )
                            )
                    elif not isinstance(split_string, str):
                        message = (
                            "Split has to be of type string or valid function for {0}"
                        )
                        matches.append(
                            RuleMatch(tree, message.format("/".join(map(str, tree))))
                        )
                else:
                    message = "Split should be an array of 2 for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree))))
                    )
            else:
                message = "Split should be an array of 2 for {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class BackupPlanLifecycleRule(CloudFormationLintRule):
    """Check Backup Plan rules with lifecycle has minimum period between cold and delete"""

    id = "E3504"
    shortdesc = "Check minimum 90 period is met between BackupPlan cold and delete"
    description = "Check that Backup plans with lifecycle rules have >= 90 days between cold and delete"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-backup-backupplan-lifecycleresourcetype.html"
    tags = ["properties", "backup", "plan", "lifecycle"]

    def match(self, cfn):
        """Check cold storage and deletion lifecycle period differences"""
        matches = []
        results = cfn.get_resource_properties(
            ["AWS::Backup::BackupPlan", "BackupPlan", "BackupPlanRule", "Lifecycle"]
        )

        for result in results:
            backup_rule = result["Value"]
            # if 'MoveToColdStorageAfterDays' in backup_rule and 'DeleteAfterDays' in backup_rule:
            if isinstance(
                backup_rule.get("MoveToColdStorageAfterDays"), int
            ) and isinstance(backup_rule.get("DeleteAfterDays"), int):
                if (
                    backup_rule["DeleteAfterDays"]
                    - backup_rule["MoveToColdStorageAfterDays"]
                    < 90
                ):
                    message = "DeleteAfterDays in {0} must be at least 90 days after MoveToColdStorageAfterDays"
                    rule_match = RuleMatch(
                        result["Path"],
                        message.format("/".join(map(str, result["Path"]))),
                    )
                    matches.append(rule_match)

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RetentionPeriodOnResourceTypesWithAutoExpiringContent(CloudFormationLintRule):
    """Check for RetentionPeriod"""

    id = "I3013"
    shortdesc = (
        "Check resources with auto expiring content have explicit retention period"
    )
    description = (
        "The behaviour for data retention is different across AWS Services."
        "If no retention period is specified the default for some services is to delete the data after a period of time."
        "This check requires you to explicitly set the retention period for those resources to avoid unexpected data losses"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources", "retentionperiod"]

    def match(self, cfn):
        """Check for RetentionPeriod"""
        matches = []

        retention_attributes_by_resource_type = {
            "AWS::Kinesis::Stream": [
                {
                    "Attribute": "RetentionPeriodHours",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-kinesis-stream.html#cfn-kinesis-stream-retentionperiodhours",
                }
            ],
            "AWS::SQS::Queue": [
                {
                    "Attribute": "MessageRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-sqs-queues.html#aws-sqs-queue-msgretentionperiod",
                }
            ],
            "AWS::DocDB::DBCluster": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-docdb-dbcluster.html#cfn-docdb-dbcluster-backupretentionperiod",
                }
            ],
            "AWS::Synthetics::Canary": [
                {
                    "Attribute": "SuccessRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-synthetics-canary.html#cfn-synthetics-canary-successretentionperiod",
                },
                {
                    "Attribute": "FailureRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-synthetics-canary.html#cfn-synthetics-canary-failureretentionperiod",
                },
            ],
            "AWS::Redshift::Cluster": [
                {
                    "Attribute": "AutomatedSnapshotRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-redshift-cluster.html#cfn-redshift-cluster-automatedsnapshotretentionperiod",
                }
            ],
            "AWS::RDS::DBInstance": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-rds-database-instance.html#cfn-rds-dbinstance-backupretentionperiod",
                    "CheckAttribute": "Engine",
                    "CheckAttributeRegex": re.compile("^((?!aurora).)*$"),
                }
            ],
            "AWS::RDS::DBCluster": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-rds-dbcluster.html#cfn-rds-dbcluster-backuprententionperiod",
                }
            ],
        }

        resources = cfn.get_resources()
        for r_name, r_values in resources.items():
            if r_values.get("Type") in retention_attributes_by_resource_type:
                for attr_def in retention_attributes_by_resource_type[
                    r_values.get("Type")
                ]:
                    property_sets = r_values.get_safe("Properties")
                    for property_set, path in property_sets:
                        error_path = ["Resources", r_name] + path
                        if not property_set:
                            message = f'The default retention period will delete the data after a pre-defined time. Set an explicit values to avoid data loss on resource : {"/".join(str(x) for x in error_path)}'
                            matches.append(RuleMatch(error_path, message))
                        else:
                            value = property_set.get(attr_def.get("Attribute"))
                            if not value:
                                message = f'The default retention period will delete the data after a pre-defined time. Set an explicit values to avoid data loss on resource : {"/".join(str(x) for x in error_path)}'
                                if attr_def.get("CheckAttribute"):
                                    if self._validate_property(
                                        property_set.get(
                                            attr_def.get("CheckAttribute")
                                        ),
                                        attr_def.get("CheckAttributeRegex"),
                                    ):
                                        matches.append(RuleMatch(error_path, message))
                                else:
                                    matches.append(RuleMatch(error_path, message))
                            if isinstance(value, dict):
                                # pylint: disable=protected-access
                                refs = cfn._search_deep_keys(
                                    "Ref",
                                    value,
                                    error_path + [attr_def.get("Attribute")],
                                )
                                for ref in refs:
                                    if ref[-1] == "AWS::NoValue":
                                        message = f'The default retention period will delete the data after a pre-defined time. Set an explicit values to avoid data loss on resource : {"/".join(str(x) for x in ref[0:-1])}'
                                        matches.append(RuleMatch(ref[0:-1], message))

        return matches

    def _validate_property(self, value, regex) -> bool:
        if isinstance(value, str):
            if regex.match(value):
                return True
            return False
        return True



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Config(CloudFormationLintRule):
    """Check if Metadata configuration is properly configured"""

    id = "E4002"
    shortdesc = "Validate the configuration of the Metadata section"
    description = "Validates that Metadata section is an object and has no null values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/metadata-section-structure.html"
    tags = ["metadata"]

    def _check_object(self, obj, path):
        results = []
        if isinstance(obj, (dict)):
            for k, v in obj.items():
                results.extend(self._check_object(v, path + [k]))
        if isinstance(obj, (list)):
            for i, v in enumerate(obj):
                results.extend(self._check_object(v, path + [i]))
        if obj is None:
            message = "Metadata value cannot be null"
            results.append(RuleMatch(path, message.format(message)))

        return results

    def match(self, cfn):
        """Check CloudFormation Metadata Interface Configuration"""

        matches = []

        metadata_obj = cfn.template.get("Metadata", {})
        if metadata_obj is None:
            message = "Metadata value has to be an object"
            matches.append(RuleMatch(["Metadata"], message.format(message)))

        if metadata_obj:
            if isinstance(metadata_obj, dict):
                matches.extend(self._check_object(metadata_obj, ["Metadata"]))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class UniqueNames(CloudFormationLintRule):
    id = "E3007"
    shortdesc = "Unique resource and parameter names"
    description = "All resources and parameters must have unique names"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/resources-section-structure.html"
    tags = ["parameters", "resources"]

    def match(self, cfn):
        matches = []
        for resource in cfn.get_resources():
            if resource in cfn.template.get("Parameters", {}):
                matches.append(
                    RuleMatch(
                        ["Resources", resource],
                        "Resources and Parameters must not share name: " + resource,
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import hashlib
import json

from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ListDuplicatesAllowed(CloudFormationLintRule):
    """Check if duplicates exist in a List"""

    id = "I3037"
    shortdesc = "Check if a list that allows duplicates has any duplicates"
    description = (
        "Certain lists support duplicate items."
        "Provide an alert when list of strings or numbers have repeats."
    )
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/rules.md#rules-1"
    tags = ["resources", "property", "list"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def _check_duplicates(self, values, path, scenario=None):
        """Check for Duplicates"""
        matches = []

        list_items = []
        if isinstance(values, list):
            for index, value in enumerate(values):
                value_hash = hashlib.sha1(
                    json.dumps(value, sort_keys=True).encode("utf-8")
                ).hexdigest()
                if value_hash in list_items:
                    if not scenario:
                        message = "List has a duplicate value at {0}"
                        matches.append(
                            RuleMatch(
                                path + [index],
                                message.format("/".join(map(str, path + [index]))),
                            )
                        )
                    else:
                        scenario_text = " and ".join(
                            [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
                        )
                        message = "List has a duplicate value at {0} when {1}"
                        matches.append(
                            RuleMatch(
                                path,
                                message.format("/".join(map(str, path)), scenario_text),
                            )
                        )

                list_items.append(value_hash)

        return matches

    def check_duplicates(self, values, path, cfn):
        """Check for duplicates"""
        matches = []

        if isinstance(values, list):
            matches.extend(self._check_duplicates(values, path))
        elif isinstance(values, dict):
            props = cfn.get_object_without_conditions(values)
            for prop in props:
                matches.extend(
                    self._check_duplicates(
                        prop.get("Object"), path, prop.get("Scenario")
                    )
                )

        return matches

    def check(self, cfn, properties, value_specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in value_specs:
                    property_type = value_specs.get(prop).get("Type")
                    primitive_type = value_specs.get(prop).get("PrimitiveItemType")
                    duplicates_allowed = value_specs.get(prop).get(
                        "DuplicatesAllowed", False
                    )
                    if (
                        property_type == "List"
                        and duplicates_allowed
                        and primitive_type in ["String", "Integer"]
                    ):
                        matches.extend(
                            self.check_duplicates(p_value[prop], p_path + [prop], cfn)
                        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LambdaMemorySize(CloudFormationLintRule):
    """Check Lambda Memory Size"""

    id = "W2510"
    shortdesc = "Parameter Memory Size attributes should have max and min"
    description = (
        "Check if a parameter that is used for Lambda memory size "
        " should have a min and max size that matches Lambda constraints"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/API_CreateFunction.html#SSS-CreateFunction-request-MemorySize"
    tags = ["parameters", "lambda"]

    def __init__(self):
        """Init"""
        super().__init__()
        resource_type_specs = [
            "AWS::Lambda::Function",
        ]

        for resource_type_spec in resource_type_specs:
            self.resource_property_types.append(resource_type_spec)

    # pylint: disable=W0613
    def check_lambda_memory_size_ref(self, value, path, parameters, resources):
        matches = []

        if value in parameters:
            parameter = parameters.get(value)
            min_value = parameter.get("MinValue")
            max_value = parameter.get("MaxValue")
            allowed_values = parameter.get("AllowedValues")
            if (not min_value or not max_value) and (not allowed_values):
                param_path = ["Parameters", value]
                message = (
                    "Lambda Memory Size parameters should use MinValue, MaxValue "
                    "or AllowedValues at {0}"
                )
                matches.append(
                    RuleMatch(param_path, message.format(("/".join(param_path))))
                )

        return matches

    def check(self, properties, resource_type, path, cfn):
        """Check itself"""
        matches = []

        matches.extend(
            cfn.check_value(
                properties,
                "MemorySize",
                path,
                check_value=None,
                check_ref=self.check_lambda_memory_size_ref,
                check_find_in_map=None,
                check_split=None,
                check_join=None,
            )
        )

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        matches.extend(self.check(properties, resource_type, path, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from cfnlint.data import AdditionalSpecs
from cfnlint.helpers import convert_dict, load_resource
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Permissions(CloudFormationLintRule):
    """Check IAM Permission configuration"""

    id = "W3037"
    shortdesc = "Check IAM Permission configuration"
    description = "Check for valid IAM Permissions"
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_action.html"
    tags = ["properties", "iam", "permissions"]
    experimental = True

    IAM_PERMISSION_RESOURCE_TYPES = {
        "AWS::Elasticsearch::Domain": "AccessPolicies",
        "AWS::OpenSearchService::Domain": "AccessPolicies",
        "AWS::IAM::Group": "Policies",
        "AWS::IAM::ManagedPolicy": "PolicyDocument",
        "AWS::IAM::Policy": "PolicyDocument",
        "AWS::IAM::Role": "Policies",
        "AWS::IAM::User": "Policies",
        "AWS::KMS::Key": "KeyPolicy",
        "AWS::S3::BucketPolicy": "PolicyDocument",
        "AWS::SNS::TopicPolicy": "PolicyDocument",
        "AWS::SQS::QueuePolicy": "PolicyDocument",
    }

    def __init__(self):
        """Init"""
        super().__init__()
        self.service_map = self.load_service_map()
        for resource_type in self.IAM_PERMISSION_RESOURCE_TYPES:
            self.resource_property_types.append(resource_type)

    def load_service_map(self):
        """
        Convert policies.json into a simpler version for more efficient key lookup.
        """
        service_map = load_resource(AdditionalSpecs, "Policies.json")["serviceMap"]

        policy_service_map = {}

        for _, properties in service_map.items():
            # The services and actions are case insensitive
            service = properties["StringPrefix"].lower()
            actions = [x.lower() for x in properties["Actions"]]

            # Some services have the same name for different generations; like elasticloadbalancing.
            if service in policy_service_map:
                policy_service_map[service] += actions
            else:
                policy_service_map[service] = actions

        return policy_service_map

    def check_policy_document(self, value, path, start_mark, end_mark):
        """Check policy document"""
        matches = []

        if isinstance(value, str):
            try:
                value = convert_dict(json.loads(value), start_mark, end_mark)
            except Exception as ex:  # pylint: disable=W0703,W0612
                message = "IAM Policy Documents need to be JSON"
                matches.append(RuleMatch(path[:], message))
                return matches

        if not isinstance(value, dict):
            return matches

        for p_vs, p_p in value.items_safe(path[:], (dict)):
            statements = p_vs.get("Statement")
            if statements:
                if isinstance(statements, dict):
                    statements = [statements]
                if isinstance(statements, list):
                    for index, statement in enumerate(statements):
                        actions = []

                        if isinstance(statement, dict):
                            actions.extend(self.get_actions(statement))

                        elif isinstance(statement, list):
                            for effective_permission in statement:
                                actions.extend(self.get_actions(effective_permission))

                        for action in actions:
                            matches.extend(
                                self.check_permissions(
                                    action, p_p + ["Statement", index]
                                )
                            )

        return matches

    def check_permissions(self, action, path):
        """
        Check if permission is valid
        """
        matches = []
        if action == "*":
            return matches

        service, permission = action.split(":", 1)
        # Get lowercase so we can check case insenstive. Keep the original values for the message
        service_value = service.lower()
        permission_value = permission.lower()

        if service_value in self.service_map:
            if permission_value == "*":
                pass
            elif permission_value.endswith("*"):
                wilcarded_permission = permission_value.split("*")[0]
                if not any(
                    wilcarded_permission in action
                    for action in self.service_map[service_value]
                ):
                    message = 'Invalid permission "{}" for "{}" found in permissions'
                    matches.append(RuleMatch(path, message.format(permission, service)))

            elif permission_value.startswith("*"):
                wilcarded_permission = permission_value.split("*")[1]
                if not any(
                    wilcarded_permission in action
                    for action in self.service_map[service_value]
                ):
                    message = 'Invalid permission "{}" for "{}" found in permissions'
                    matches.append(RuleMatch(path, message.format(permission, service)))
            elif permission_value not in self.service_map[service_value]:
                message = 'Invalid permission "{}" for "{}" found in permissions'
                matches.append(RuleMatch(path, message.format(permission, service)))
        else:
            message = 'Invalid service "{}" found in permissions'
            matches.append(RuleMatch(path, message.format(service)))

        return matches

    def get_actions(self, effective_permissions):
        """return all actions from a statement"""

        actions = []

        if "Action" in effective_permissions:
            if isinstance(effective_permissions.get("Action"), str):
                actions.append(effective_permissions.get("Action"))
            elif isinstance(effective_permissions.get("Action"), list):
                actions.extend(effective_permissions.get("Action"))

        if "NotAction" in effective_permissions:
            if isinstance(effective_permissions.get("NotAction"), str):
                actions.append(effective_permissions.get("NotAction"))
            elif isinstance(effective_permissions.get("Action"), list):
                actions.extend(effective_permissions.get("NotAction"))

        return actions

    def match_resource_properties(self, properties, resourcetype, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        key = self.IAM_PERMISSION_RESOURCE_TYPES.get(resourcetype)
        for key, value in properties.items():
            if key == "Policies" and isinstance(value, list):
                for index, policy in enumerate(properties.get(key, [])):
                    matches.extend(
                        cfn.check_value(
                            obj=policy,
                            key="PolicyDocument",
                            path=path[:] + ["Policies", index],
                            check_value=self.check_policy_document,
                            start_mark=key.start_mark,
                            end_mark=key.end_mark,
                        )
                    )
            elif key in [
                "KeyPolicy",
                "PolicyDocument",
                "RepositoryPolicyText",
                "AccessPolicies",
            ]:
                matches.extend(
                    cfn.check_value(
                        obj=properties,
                        key=key,
                        path=path[:],
                        check_value=self.check_policy_document,
                        start_mark=key.start_mark,
                        end_mark=key.end_mark,
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LimitDescription(CloudFormationLintRule):
    """Check Template Description Size"""

    id = "E1003"
    shortdesc = "Template description limit"
    description = (
        "Check if the size of the template description is less than the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["description", "limits"]

    def match(self, cfn):
        matches = []
        description = cfn.template.get("Description", "")
        if isinstance(description, str):
            if len(description) > LIMITS["template"]["description"]:
                message = (
                    "The template description ({0} bytes) exceeds the limit ({1} bytes)"
                )
                matches.append(
                    RuleMatch(
                        ["Description"],
                        message.format(
                            len(description), LIMITS["template"]["description"]
                        ),
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import PSEUDOPARAMS, VALID_PARAMETER_TYPES_LIST
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Sub(CloudFormationLintRule):
    """Check if Sub values are correct"""

    id = "E1019"
    shortdesc = "Sub validation of parameters"
    description = "Making sure the sub function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def _test_string(self, cfn, sub_string, parameters, tree):
        """Test if a string has appropriate parameters"""

        matches = []
        string_params = cfn.get_sub_parameters(sub_string)

        for string_param in string_params:
            if isinstance(string_param, (str)):
                matches.extend(
                    self._test_parameter(string_param, cfn, parameters, tree)
                )

        return matches

    def _get_parameters(self, cfn):
        """Get all Parameter Names"""
        results = {}
        parameters = cfn.template.get("Parameters", {})
        if isinstance(parameters, dict):
            for param_name, param_values in parameters.items():
                # This rule isn't here to check the Types but we need
                # something valid if it doesn't exist
                if isinstance(param_values, dict):
                    results[param_name] = param_values.get("Type", "String")

        return results

    def _test_parameters(self, parameters, cfn, tree):
        """Check parameters for appropriate configuration"""

        supported_functions = [
            "Fn::Base64",
            "Fn::FindInMap",
            "Fn::GetAZs",
            "Fn::GetAtt",
            "Fn::If",
            "Fn::ImportValue",
            "Fn::Join",
            "Fn::Select",
            "Fn::Sub",
            "Ref",
            "Fn::ToJsonString",
        ]

        matches = []
        for parameter_name, parameter_value_obj in parameters.items():
            param_tree = tree[:] + [parameter_name]
            if isinstance(parameter_value_obj, dict):
                if len(parameter_value_obj) == 1:
                    for key, value in parameter_value_obj.items():
                        if key not in supported_functions:
                            message = (
                                "Sub parameter should use a valid function for {0}"
                            )
                            matches.append(
                                RuleMatch(
                                    param_tree, message.format("/".join(map(str, tree)))
                                )
                            )
                        elif key in ["Ref"]:
                            matches.extend(self._test_parameter(value, cfn, {}, tree))
                        elif key in ["Fn::GetAtt"]:
                            if isinstance(value, list):
                                # Only test this if all the items are a string
                                if_all_strings = True
                                for v in value:
                                    if not isinstance(v, str):
                                        # skip things got too complex
                                        if_all_strings = False
                                if if_all_strings:
                                    matches.extend(
                                        self._test_parameter(
                                            ".".join(value), cfn, {}, tree
                                        )
                                    )
                            elif isinstance(value, str):
                                matches.extend(
                                    self._test_parameter(value, cfn, {}, tree)
                                )
                else:
                    message = "Sub parameter should be an object of 1 for {0}"
                    matches.append(
                        RuleMatch(param_tree, message.format("/".join(map(str, tree))))
                    )
            elif isinstance(parameter_value_obj, list):
                message = "Sub parameter value should be a string for {0}"
                matches.append(
                    RuleMatch(param_tree, message.format("/".join(map(str, tree))))
                )

        return matches

    def _test_parameter(self, parameter, cfn, parameters, tree):
        """Test a parameter"""

        matches = []
        get_atts = cfn.get_valid_getatts()

        valid_params = list(PSEUDOPARAMS)
        valid_params.extend(cfn.get_resource_names())
        template_parameters = self._get_parameters(cfn)

        for key, _ in parameters.items():
            valid_params.append(key)

        if parameter not in valid_params:
            found = False
            if parameter in template_parameters:
                found = True
                if template_parameters.get(parameter) in VALID_PARAMETER_TYPES_LIST:
                    message = "Fn::Sub cannot use list {0} at {1}"
                    matches.append(
                        RuleMatch(
                            tree, message.format(parameter, "/".join(map(str, tree)))
                        )
                    )
            for resource, attributes in get_atts.items():
                for attribute_name, attribute_values in attributes.items():
                    if resource == parameter.split(".")[0]:
                        if attribute_name == "*":
                            found = True
                        elif attribute_name == ".".join(parameter.split(".")[1:]):
                            if attribute_values.get("Type") == "List":
                                message = "Fn::Sub cannot use list {0} at {1}"
                                matches.append(
                                    RuleMatch(
                                        tree,
                                        message.format(
                                            parameter, "/".join(map(str, tree))
                                        ),
                                    )
                                )
                            found = True
                        else:
                            if (
                                attribute_name == parameter.split(".")[1]
                                and attribute_values.get("Type") == "Map"
                            ):
                                found = True

            if not found:
                message = "Parameter {0} for Fn::Sub not found at {1}"
                matches.append(
                    RuleMatch(tree, message.format(parameter, "/".join(map(str, tree))))
                )

        return matches

    def match(self, cfn):
        matches = []

        sub_objs = cfn.search_deep_keys("Fn::Sub")

        for sub_obj in sub_objs:
            sub_value_obj = sub_obj[-1]
            tree = sub_obj[:-1]
            if isinstance(sub_value_obj, str):
                matches.extend(self._test_string(cfn, sub_value_obj, {}, tree))
            elif isinstance(sub_value_obj, list):
                if len(sub_value_obj) == 2:
                    sub_string = sub_value_obj[0]
                    parameters = sub_value_obj[1]
                    if not isinstance(sub_string, str):
                        message = "Subs first element should be of type string for {0}"
                        matches.append(
                            RuleMatch(
                                tree + [0], message.format("/".join(map(str, tree)))
                            )
                        )
                    if not isinstance(parameters, dict):
                        message = "Subs second element should be an object for {0}"
                        matches.append(
                            RuleMatch(
                                tree + [1], message.format("/".join(map(str, tree)))
                            )
                        )
                    else:
                        matches.extend(
                            self._test_string(cfn, sub_string, parameters, tree + [0])
                        )
                        matches.extend(self._test_parameters(parameters, cfn, tree))
                else:
                    message = "Sub should be an array of 2 for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree))))
                    )
            elif isinstance(sub_value_obj, dict):
                if len(sub_value_obj) == 1:
                    for key, _ in sub_value_obj.items():
                        if not key == "Fn::Transform":
                            message = (
                                "Sub should be a string or array of 2 items for {0}"
                            )
                            matches.append(
                                RuleMatch(
                                    tree, message.format("/".join(map(str, tree)))
                                )
                            )
                else:
                    message = "Sub should be a string or array of 2 items for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree))))
                    )
            else:
                message = "Sub should be a string or array of 2 items for {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.helpers import LIMITS, REGEX_ALPHANUMERIC
from cfnlint.rules import RuleMatch


def approaching_name_limit(cfn, section):
    matches = []
    for name in cfn.template.get(section, {}):
        if (
            LIMITS["threshold"] * LIMITS[section]["name"]
            < len(name)
            <= LIMITS[section]["name"]
        ):
            message = (
                "The length of "
                + section[:-1]
                + " name ({0}) is approaching the limit ({1})"
            )
            matches.append(
                RuleMatch(
                    [section, name], message.format(len(name), LIMITS[section]["name"])
                )
            )
    return matches


def approaching_number_limit(cfn, section):
    matches = []
    number = (
        cfn.get_resources() if section == "Resources" else cfn.template.get(section, {})
    )
    if (
        LIMITS["threshold"] * LIMITS[section]["number"]
        < len(number)
        <= LIMITS[section]["number"]
    ):
        message = "The number of " + section + " ({0}) is approaching the limit ({1})"
        matches.append(
            RuleMatch([section], message.format(len(number), LIMITS[section]["number"]))
        )
    return matches


def name_limit(cfn, section):
    matches = []
    for name in cfn.template.get(section, {}):
        if len(name) > LIMITS[section]["name"]:
            message = (
                "The length of " + section[:-1] + " name ({0}) exceeds the limit ({1})"
            )
            matches.append(
                RuleMatch(
                    [section, name], message.format(len(name), LIMITS[section]["name"])
                )
            )
    return matches


def number_limit(cfn, section):
    matches = []
    number = (
        cfn.get_resources() if section == "Resources" else cfn.template.get(section, {})
    )
    if len(number) > LIMITS[section]["number"]:
        message = "The number of " + section + " ({0}) exceeds the limit ({1})"
        matches.append(
            RuleMatch([section], message.format(len(number), LIMITS[section]["number"]))
        )
    return matches


def alphanumeric_name(cfn, section):
    matches = []
    for name, _ in cfn.template.get(section, {}).items():
        if not re.match(REGEX_ALPHANUMERIC, name):
            message = (
                section[:-1] + " {0} has invalid name.  Name has to be alphanumeric."
            )
            matches.append(RuleMatch([section, name], message.format(name)))
    return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Ebs(CloudFormationLintRule):
    """Check Ec2 Ebs Resource Properties"""

    id = "E2504"
    shortdesc = "Check Ec2 Ebs Properties"
    description = "See if Ec2 Ebs Properties are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-blockdev-template.html"
    tags = ["properties", "ec2", "ebs"]

    def _checkEbs(self, cfn, ebs, path):
        matches = []

        if isinstance(ebs, dict):
            volume_types_obj = cfn.get_values(ebs, "VolumeType")
            iops_obj = cfn.get_values(ebs, "Iops")
            if volume_types_obj is not None:
                for volume_type_obj in volume_types_obj:
                    volume_type = volume_type_obj.get("Value")
                    if isinstance(volume_type, str):
                        if volume_type in ("io1", "io2"):
                            if iops_obj is None:
                                pathmessage = path[:] + ["VolumeType"]
                                message = "VolumeType {0} requires Iops to be specified for {1}"
                                matches.append(
                                    RuleMatch(
                                        pathmessage,
                                        message.format(
                                            volume_type, "/".join(map(str, pathmessage))
                                        ),
                                    )
                                )
                        elif volume_type in ("gp2", "st1", "sc1", "standard"):
                            if iops_obj is not None:
                                pathmessage = path[:] + ["Iops"]
                                message = (
                                    "Iops shouldn't be defined for type {0} for {1}"
                                )
                                matches.append(
                                    RuleMatch(
                                        pathmessage,
                                        message.format(
                                            volume_type, "/".join(map(str, pathmessage))
                                        ),
                                    )
                                )

        return matches

    def match(self, cfn):
        """Check Ec2 Ebs Resource Parameters"""

        matches = []

        results = cfn.get_resource_properties(
            ["AWS::EC2::Instance", "BlockDeviceMappings"]
        )
        results.extend(
            cfn.get_resource_properties(
                ["AWS::AutoScaling::LaunchConfiguration", "BlockDeviceMappings"]
            )
        )
        for result in results:
            path = result["Path"]
            if isinstance(result["Value"], list):
                for index, properties in enumerate(result["Value"]):
                    virtual_name = properties.get("VirtualName")
                    ebs = properties.get("Ebs")
                    if virtual_name:
                        # switch to regex
                        if not re.match(
                            r"^ephemeral([0-9]|[1][0-9]|[2][0-3])$", virtual_name
                        ):
                            pathmessage = path[:] + [index, "VirtualName"]
                            message = "Property VirtualName should be of type ephemeral(n) for {0}"
                            matches.append(
                                RuleMatch(
                                    pathmessage,
                                    message.format("/".join(map(str, pathmessage))),
                                )
                            )
                    elif ebs:
                        matches.extend(
                            self._checkEbs(cfn, ebs, path[:] + [index, "Ebs"])
                        )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from cfnlint.helpers import REGEX_CIDR
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Cidr(CloudFormationLintRule):
    """Check if Cidr values are correct"""

    id = "E1024"
    shortdesc = "Cidr validation of parameters"
    description = "Making sure the function CIDR is a list with valid values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-cidr.html"
    tags = ["functions", "cidr"]

    supported_functions = [
        "Fn::FindInMap",
        "Fn::Select",
        "Ref",
        "Fn::GetAtt",
        "Fn::Sub",
        "Fn::ImportValue",
    ]

    def check_ip_block(self, value, path):
        matches = []
        if isinstance(value, dict):
            if len(value) == 1:
                for index_key, _ in value.items():
                    if index_key not in self.supported_functions:
                        if index_key == "Fn::If":
                            if len(value.get("Fn::If")) == 3 and isinstance(
                                value.get("Fn::If"), list
                            ):
                                matches.extend(
                                    self.check_ip_block(
                                        value.get("Fn::If")[1],
                                        path=path[:] + [index_key, 1],
                                    )
                                )
                                matches.extend(
                                    self.check_ip_block(
                                        value.get("Fn::If")[2],
                                        path=path[:] + [index_key, 2],
                                    )
                                )
                        else:
                            message = "Cidr ipBlock should be Cidr Range, Ref, GetAtt, Sub or Select for {0}"
                            matches.append(
                                RuleMatch(
                                    path, message.format("/".join(map(str, value)))
                                )
                            )
        elif isinstance(value, (str)):
            if not re.match(REGEX_CIDR, value):
                message = "Cidr ipBlock should be a Cidr Range based string for {0}"
                matches.append(
                    RuleMatch(path, message.format("/".join(map(str, path))))
                )
        else:
            message = "Cidr ipBlock should be a string for {0}"
            matches.append(RuleMatch(path, message.format("/".join(map(str, path)))))

        return matches

    def check_count(self, value, path):
        matches = []
        count_parameters = []
        if isinstance(value, dict):
            if len(value) == 1:
                for index_key, index_value in value.items():
                    if index_key not in self.supported_functions:
                        if index_key == "Fn::If":
                            if len(value.get("Fn::If")) == 3 and isinstance(
                                value.get("Fn::If"), list
                            ):
                                matches.extend(
                                    self.check_count(
                                        value.get("Fn::If")[1],
                                        path=path[:] + [index_key, 1],
                                    )
                                )
                                matches.extend(
                                    self.check_count(
                                        value.get("Fn::If")[2],
                                        path=path[:] + [index_key, 2],
                                    )
                                )
                        else:
                            message = "Cidr count should be Int, Ref, or Select for {0}"
                            matches.append(
                                RuleMatch(
                                    path, message.format("/".join(map(str, path)))
                                )
                            )
                    if index_key == "Ref":
                        count_parameters.append(index_value)
        elif not isinstance(value, int):
            message = "Cidr count should be a int for {0}"
            extra_args = {
                "actual_type": type(value).__name__,
                "expected_type": int.__name__,
            }
            matches.append(
                RuleMatch(path, message.format("/".join(map(str, path))), **extra_args)
            )

        return count_parameters, matches

    def check_size_mask(self, value, path):
        matches = []
        size_mask_parameters = []
        if isinstance(value, dict):
            if len(value) == 1:
                for index_key, index_value in value.items():
                    if index_key not in self.supported_functions:
                        if index_key == "Fn::If":
                            if len(value.get("Fn::If")) == 3 and isinstance(
                                value.get("Fn::If"), list
                            ):
                                matches.extend(
                                    self.check_size_mask(
                                        value.get("Fn::If")[1],
                                        path=path[:] + [index_key, 1],
                                    )
                                )
                                matches.extend(
                                    self.check_size_mask(
                                        value.get("Fn::If")[2],
                                        path=path[:] + [index_key, 2],
                                    )
                                )
                        else:
                            message = (
                                "Cidr sizeMask should be Int, Ref, or Select for {0}"
                            )
                            matches.append(
                                RuleMatch(
                                    path, message.format("/".join(map(str, path)))
                                )
                            )
                    if index_key == "Ref":
                        size_mask_parameters.append(index_value)
        elif not isinstance(value, int):
            message = "Cidr sizeMask should be a int for {0}"
            extra_args = {
                "actual_type": type(value).__name__,
                "expected_type": int.__name__,
            }
            matches.append(
                RuleMatch(path, message.format("/".join(map(str, path))), **extra_args)
            )

        return size_mask_parameters, matches

    def check_parameter_count(self, cfn, parameter_name):
        """Check Count Parameter if used"""
        matches = []
        parameter_obj = cfn.get_parameters().get(parameter_name, {})
        if parameter_obj:
            tree = ["Parameters", parameter_name]
            parameter_type = parameter_obj.get("Type")
            if parameter_type == "Number":
                max_value = parameter_obj.get("MaxValue")
                min_value = parameter_obj.get("MinValue")
                if (not min_value) or min_value < 1 or min_value > 256:
                    message = "Parameter for Cidr count have MinValue between 1 and 256 at {0}"
                    matches.append(
                        RuleMatch(
                            tree + ["MinValue"],
                            message.format("/".join(map(str, tree + ["MinValue"]))),
                        )
                    )
                if (not max_value) or max_value < 1 or max_value > 256:
                    message = "Parameter for Cidr count have MaxValue between 1 and 256 at {0}"
                    matches.append(
                        RuleMatch(
                            tree + ["MaxValue"],
                            message.format("/".join(map(str, tree + ["MaxValue"]))),
                        )
                    )
            else:
                message = "Parameter for Cidr count have be of Type Number at {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )

        return matches

    def check_parameter_size_mask(self, cfn, parameter_name):
        """Check SizeMask Parameter if used"""
        matches = []
        parameter_obj = cfn.get_parameters().get(parameter_name, {})
        if parameter_obj:
            tree = ["Parameters", parameter_name]
            parameter_type = parameter_obj.get("Type")
            if parameter_type == "Number":
                max_value = parameter_obj.get("MaxValue")
                min_value = parameter_obj.get("MinValue")
                if (not min_value) or min_value < 1 or min_value > 256:
                    message = (
                        "Parameter for Cidr sizeMask have MinValue between 1 and "
                        "128 (for ipv6) and 32 (for ipv4) at {0}"
                    )
                    matches.append(
                        RuleMatch(
                            tree + ["MinValue"],
                            message.format("/".join(map(str, tree + ["MinValue"]))),
                        )
                    )
                if (not max_value) or max_value < 1 or max_value > 256:
                    message = (
                        "Parameter for Cidr count have MaxValue between 1 and "
                        "128 (for ipv6) and 32 (for ipv4) at {0}"
                    )
                    matches.append(
                        RuleMatch(
                            tree + ["MaxValue"],
                            message.format("/".join(map(str, tree + ["MaxValue"]))),
                        )
                    )
            else:
                message = "Parameter for Cidr count have be of Type Number at {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )

        return matches

    def match(self, cfn):
        matches = []

        cidr_objs = cfn.search_deep_keys("Fn::Cidr")

        count_parameters = []
        size_mask_parameters = []

        for cidr_obj in cidr_objs:
            cidr_value_obj = cidr_obj[-1]
            tree = cidr_obj[:-1]
            if isinstance(cidr_value_obj, list):
                if len(cidr_value_obj) in [2, 3]:
                    ip_block_obj = cidr_value_obj[0]
                    count_obj = cidr_value_obj[1]
                    if len(cidr_value_obj) == 3:
                        size_mask_obj = cidr_value_obj[2]
                    else:
                        size_mask_obj = None

                    matches.extend(self.check_ip_block(ip_block_obj, tree[:] + [0]))

                    new_count_parameters, new_matches = self.check_count(
                        count_obj, tree[:] + [1]
                    )
                    count_parameters.extend(new_count_parameters)
                    matches.extend(new_matches)

                    new_size_mask_parameters, new_matches = self.check_size_mask(
                        size_mask_obj, tree[:] + [2]
                    )
                    size_mask_parameters.extend(new_size_mask_parameters)
                    matches.extend(new_matches)

                else:
                    message = "Cidr should be a list of 2 or 3 elements for {0}"
                    matches.append(
                        RuleMatch(tree, message.format("/".join(map(str, tree))))
                    )
            else:
                message = "Cidr should be a list of 2 or 3 elements for {0}"
                matches.append(
                    RuleMatch(tree, message.format("/".join(map(str, tree))))
                )

        for count_parameter in set(count_parameters):
            matches.extend(self.check_parameter_count(cfn, count_parameter))
        for size_mask_parameter in set(size_mask_parameters):
            matches.extend(self.check_parameter_size_mask(cfn, size_mask_parameter))

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
from typing import List, Optional, Sequence, Union

from cfnlint.graph import Graph
from cfnlint.rules import Match, RulesCollection
from cfnlint.template import Template
from cfnlint.transform import Transform

LOGGER = logging.getLogger(__name__)


class Runner:
    """Run all the rules"""

    def __init__(
        self,
        rules: RulesCollection,
        filename: Optional[str],
        template: str,
        regions: Sequence[str],
        verbosity=0,
        mandatory_rules: Union[Sequence[str], None] = None,
    ):
        self.rules = rules
        self.filename = filename
        self.verbosity = verbosity
        self.mandatory_rules = mandatory_rules or []
        self.cfn = Template(filename, template, regions)

    def transform(self):
        """Transform logic"""
        LOGGER.debug("Transform templates if needed")
        sam_transform = "AWS::Serverless-2016-10-31"
        matches = []
        transform_declaration = self.cfn.template.get("Transform", [])
        transform_type = (
            transform_declaration
            if isinstance(transform_declaration, list)
            else [transform_declaration]
        )
        # Don't call transformation if Transform is not specified to prevent
        # useless execution of the transformation.
        # Currently locked in to SAM specific
        if sam_transform not in transform_type:
            return matches
        # Save the Globals section so its available for rule processing
        self.cfn.transform_pre["Globals"] = self.cfn.template.get("Globals", {})
        transform = Transform(self.filename, self.cfn.template, self.cfn.regions[0])
        matches = transform.transform_template()
        self.cfn.template = transform.template()
        self.cfn.graph = Graph(self.cfn)
        return matches

    def run(self) -> List[Match]:
        """Run rules"""
        LOGGER.info("Run scan of template %s", self.filename)
        matches = []
        if self.cfn.template is not None:
            matches.extend(self.rules.run(self.filename, self.cfn))
        return self.check_metadata_directives(matches)

    def check_metadata_directives(self, matches: Sequence[Match]) -> List[Match]:
        # uniq the list of incidents and filter out exceptions from the template
        directives = self.cfn.get_directives()
        return_matches: List[Match] = []
        for match in matches:
            if not any(match == u for u in return_matches):
                if match.rule.id not in directives:
                    return_matches.append(match)
                else:
                    for mandatory_rule in self.mandatory_rules:
                        if match.rule.id.startswith(mandatory_rule):
                            return_matches.append(match)
                            break
                    else:
                        for directive in directives.get(match.rule.id):
                            start = directive.get("start")
                            end = directive.get("end")
                            if start[0] < match.linenumber < end[0]:
                                break
                            if (
                                start[0] == match.linenumber
                                and start[1] <= match.columnnumber
                            ):
                                break
                            if (
                                end[0] == match.linenumber
                                and end[1] >= match.columnnumberend
                            ):
                                break
                        else:
                            return_matches.append(match)

        return return_matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.common import name_limit


class LimitName(CloudFormationLintRule):
    """Check if maximum Output name size limit is exceeded"""

    id = "E6011"
    shortdesc = "Output name limit not exceeded"
    description = (
        "Check the size of Output names in the template is less than the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["outputs", "limits"]

    def match(self, cfn):
        return name_limit(cfn, "Outputs")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SecurityGroupIngress(CloudFormationLintRule):
    """Check if EC2 Security Group Ingress Properties"""

    id = "E2506"
    shortdesc = "Resource EC2 Security Group Ingress Properties"
    description = (
        "See if EC2 Security Group Ingress Properties are set correctly. "
        'Check that "SourceSecurityGroupId" or "SourceSecurityGroupName" are '
        " are exclusive and using the type of Ref or GetAtt "
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-security-group-ingress.html"
    tags = ["resources", "ec2", "securitygroup"]

    def check_ingress_rule(self, vpc_id, properties, path):
        """Check ingress rule"""

        matches = []
        if vpc_id:
            # Check that SourceSecurityGroupName isn't specified
            if properties.get("SourceSecurityGroupName", None):
                path_error = path[:] + ["SourceSecurityGroupName"]
                message = (
                    "SourceSecurityGroupName shouldn't be specified for "
                    "Vpc Security Group at {0}"
                )
                matches.append(
                    RuleMatch(
                        path_error, message.format("/".join(map(str, path_error)))
                    )
                )

        else:
            if properties.get("SourceSecurityGroupId", None):
                path_error = path[:] + ["SourceSecurityGroupId"]
                message = (
                    "SourceSecurityGroupId shouldn't be specified for "
                    "Non-Vpc Security Group at {0}"
                )
                matches.append(
                    RuleMatch(
                        path_error, message.format("/".join(map(str, path_error)))
                    )
                )

        return matches

    def match(self, cfn):
        """Check EC2 Security Group Ingress Resource Parameters"""

        matches = []

        resources = cfn.get_resources(resource_type="AWS::EC2::SecurityGroup")
        for resource_name, resource_object in resources.items():
            properties = resource_object.get("Properties", {})
            if properties:
                vpc_id = properties.get("VpcId", None)
                ingress_rules = properties.get("SecurityGroupIngress")
                if isinstance(ingress_rules, list):
                    for index, ingress_rule in enumerate(ingress_rules):
                        path = [
                            "Resources",
                            resource_name,
                            "Properties",
                            "SecurityGroupIngress",
                            index,
                        ]
                        matches.extend(
                            self.check_ingress_rule(
                                vpc_id=vpc_id, properties=ingress_rule, path=path
                            )
                        )

        resources = None
        resources = cfn.get_resources(resource_type="AWS::EC2::SecurityGroupIngress")
        for resource_name, resource_object in resources.items():
            properties = resource_object.get("Properties", {})
            group_id = properties.get("GroupId", None)
            path = ["Resources", resource_name, "Properties"]
            if group_id:
                vpc_id = "vpc-1234567"
            else:
                vpc_id = None

            if properties:
                path = ["Resources", resource_name, "Properties"]
                matches.extend(
                    self.check_ingress_rule(
                        vpc_id=vpc_id, properties=properties, path=path
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

from jsonschema import ValidationError, validate

from cfnlint.helpers import (
    FN_PREFIX,
    PSEUDOPARAMS,
    REGEX_DYN_REF,
    REGISTRY_SCHEMAS,
    UNCONVERTED_SUFFIXES,
)
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ResourceSchema(CloudFormationLintRule):
    id = "E3000"
    shortdesc = "Resource schema"
    description = "CloudFormation Registry resource schema validation"
    source_url = (
        "https://github.com/aws-cloudformation/aws-cloudformation-resource-schema/"
    )
    tags = ["resources"]

    def match(self, cfn):
        matches = []
        for schema in REGISTRY_SCHEMAS:
            resource_type = schema["typeName"]
            for resource_name, resource_values in cfn.get_resources(
                [resource_type]
            ).items():
                properties = resource_values.get("Properties", {})
                # ignoring resources with CloudFormation template syntax in Properties
                if (
                    not re.match(REGEX_DYN_REF, str(properties))
                    and not any(
                        x in str(properties)
                        for x in PSEUDOPARAMS + UNCONVERTED_SUFFIXES
                    )
                    and FN_PREFIX not in str(properties)
                ):
                    try:
                        validate(properties, schema)
                    except ValidationError as e:
                        matches.append(
                            RuleMatch(
                                ["Resources", resource_name, "Properties"], e.message
                            )
                        )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import logging
from json.decoder import JSONDecodeError
from typing import Callable, List, Optional, Tuple, Union

from yaml import YAMLError
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from cfnlint.decode import cfn_json, cfn_yaml
from cfnlint.rules import Match, ParseError

LOGGER = logging.getLogger(__name__)

Matches = List[Match]
Decode = Tuple[Union[str, None], Matches]


def decode_str(s: str) -> Decode:
    """Decode the string s into an object."""
    return _decode(cfn_yaml.loads, cfn_json.loads, s, None)


def decode(filename: str) -> Decode:
    """Decode filename into an object."""
    return _decode(cfn_yaml.load, cfn_json.load, filename, filename)


def _decode(
    yaml_f: Callable, json_f: Callable, payload: str, filename: Optional[str]
) -> Decode:
    """Decode payload using yaml_f and json_f, using filename for log output."""
    template = None
    matches = []
    try:
        template = yaml_f(payload)
    except IOError as e:
        if e.errno == 2:
            LOGGER.error("Template file not found: %s", filename)
            matches.append(
                create_match_file_error(
                    filename, f"Template file not found: {filename}"
                )
            )
        elif e.errno == 21:
            LOGGER.error("Template references a directory, not a file: %s", filename)
            matches.append(
                create_match_file_error(
                    filename, f"Template references a directory, not a file: {filename}"
                )
            )
        elif e.errno == 13:
            LOGGER.error("Permission denied when accessing template file: %s", filename)
            matches.append(
                create_match_file_error(
                    filename,
                    f"Permission denied when accessing template file: {filename}",
                )
            )

        if matches:
            return (None, matches)
    except UnicodeDecodeError as _:
        LOGGER.error("Cannot read file contents: %s", filename)
        matches.append(
            create_match_file_error(filename, "Cannot read file contents: {filename}")
        )
    except cfn_yaml.CfnParseError as err:
        matches = err.matches
    except ParserError as err:
        matches = [create_match_yaml_parser_error(err, filename)]
    except ScannerError as err:
        if err.problem and (
            err.problem
            in [
                "found character '\\t' that cannot start any token",
                "found unknown escape character",
            ]
            or err.problem.startswith("found unknown escape character")
        ):
            try:
                template = json_f(payload)
            except cfn_json.JSONDecodeError as json_errs:
                for json_err in json_errs.matches:
                    json_err.filename = filename
                matches = json_errs.matches
            except JSONDecodeError as json_err:
                if hasattr(json_err, "msg"):
                    if json_err.msg in [
                        "No JSON object could be decoded",
                        "Expecting value",
                    ]:  # pylint: disable=no-member
                        matches = [create_match_yaml_parser_error(err, filename)]
                    else:
                        matches = [create_match_json_parser_error(json_err, filename)]
            except Exception as json_err:  # pylint: disable=W0703
                LOGGER.error("Template %s is malformed: %s", filename, err.problem)
                LOGGER.error(
                    "Tried to parse %s as JSON but got error: %s",
                    filename,
                    str(json_err),
                )
                return (
                    None,
                    [
                        create_match_file_error(
                            filename,
                            f"Tried to parse {filename} as JSON but got error: {str(json_err)}",
                        )
                    ],
                )
        else:
            matches = [create_match_yaml_parser_error(err, filename)]
    except YAMLError as err:
        matches = [create_match_file_error(filename, str(err))]

    if not isinstance(template, dict) and not matches:
        # Template isn't a dict which means nearly nothing will work
        matches = [
            Match(
                1,
                1,
                1,
                1,
                filename,
                ParseError(),
                message="Template needs to be an object.",
            )
        ]
    return (template, matches)


def create_match_yaml_parser_error(parser_error, filename):
    """Create a Match for a parser error"""
    lineno = parser_error.problem_mark.line + 1
    colno = parser_error.problem_mark.column + 1
    msg = parser_error.problem
    return Match(lineno, colno, lineno, colno + 1, filename, ParseError(), message=msg)


def create_match_file_error(filename, msg):
    """Create a Match for a parser error"""
    return Match(
        linenumber=1,
        columnnumber=1,
        linenumberend=1,
        columnnumberend=2,
        filename=filename,
        rule=ParseError(),
        message=msg,
    )


def create_match_json_parser_error(parser_error, filename):
    """Create a Match for a parser error"""
    lineno = parser_error.lineno
    colno = parser_error.colno
    msg = parser_error.msg
    return Match(lineno, colno, lineno, colno + 1, filename, ParseError(), message=msg)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class TaskDefinitionEssentialContainer(CloudFormationLintRule):
    """Check ECS TaskDefinition ContainerDefinitions Property Specifies at least one Essential Container"""

    id = "E3042"
    shortdesc = "Check at least one essential container is specified"
    description = (
        "Check that every TaskDefinition specifies at least one essential container"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ecs-taskdefinition-containerdefinitions.html#cfn-ecs-taskdefinition-containerdefinition-essential"
    tags = ["properties", "ecs", "task", "container", "fargate"]

    def match(self, cfn):
        """Check at least one essential container is specified"""

        matches = []

        results = cfn.get_resource_properties(
            ["AWS::ECS::TaskDefinition", "ContainerDefinitions"]
        )

        for result in results:
            path = result["Path"]

            has_essential_container = False

            for container in result["Value"]:
                if "Essential" in container:
                    if container["Essential"]:
                        has_essential_container = True
                    else:
                        pass
                else:
                    # If 'Essential' is not specified, it defaults to an essential container
                    has_essential_container = True

            if not has_essential_container:
                error_message = (
                    f"No essential containers defined for {'/'.join(map(str, path))}"
                )
                rule_match = RuleMatch(path, error_message)
                matches.append(rule_match)

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Required(CloudFormationLintRule):
    """Check if Outputs have required properties"""

    id = "E6002"
    shortdesc = "Outputs have required properties"
    description = "Making sure the outputs have required properties"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    def match(self, cfn):
        matches = []

        outputs = cfn.template.get("Outputs", {})
        if outputs:
            for output_name, output_value in outputs.items():
                if isinstance(output_value, dict):
                    if "Value" not in output_value:
                        message = "Output {0} is missing property {1}"
                        matches.append(
                            RuleMatch(
                                ["Outputs", output_name, "Value"],
                                message.format(output_name, "Value"),
                            )
                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import argparse
import copy
import glob
import json
import logging
import os
import sys
from pathlib import Path
from typing import Dict

import jsonschema

import cfnlint.decode.cfn_yaml
from cfnlint.helpers import REGIONS
from cfnlint.version import __version__

# pylint: disable=too-many-public-methods
LOGGER = logging.getLogger("cfnlint")


def configure_logging(debug_logging, info_logging):
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    if debug_logging:
        LOGGER.setLevel(logging.DEBUG)
    elif info_logging:
        LOGGER.setLevel(logging.INFO)
    else:
        LOGGER.setLevel(logging.NOTSET)
    log_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(log_formatter)

    # make sure all other log handlers are removed before adding it back
    for handler in LOGGER.handlers:
        LOGGER.removeHandler(handler)
    LOGGER.addHandler(ch)


class ConfigFileArgs:
    """
    Config File arguments.
    Parses .cfnlintrc OR .cfnlintrc.yaml OR .cfnlintrc.yml in the Home and Project folder.
    """

    file_args: Dict = {}
    __user_config_file = None
    __project_config_file = None
    __custom_config_file = None

    def __init__(self, schema=None, config_file=None):
        # self.file_args = self.get_config_file_defaults()
        self.file_args = {}
        self.default_schema_file = Path(__file__).parent.joinpath(
            "data/CfnLintCli/config/schema.json"
        )
        with self.default_schema_file.open(encoding="utf-8") as f:
            self.default_schema = json.load(f)
        self.schema = self.default_schema if not schema else schema

        if config_file:
            self.__custom_config_file = config_file
        else:
            LOGGER.debug("Looking for CFLINTRC before attempting to load")
            self.__user_config_file, self.__project_config_file = self._find_config()

        self.load()

    def _find_config(self):
        """Looks up for user and project level config
        Returns
        -------
        Tuple
            (Path, Path)
            Tuple with both configs and whether they were found
        Example
        -------
            > user_config, project_config = self._find_config()
        """
        config_file_name = ".cfnlintrc"
        self.__user_config_file = Path.home().joinpath(config_file_name)

        self.__project_config_file = Path.cwd().joinpath(config_file_name)
        if self._has_file(config_file_name + ".yaml"):
            self.__project_config_file = Path.cwd().joinpath(config_file_name + ".yaml")
        elif self._has_file(config_file_name + ".yml"):
            self.__project_config_file = Path.cwd().joinpath(config_file_name + ".yml")

        user_config_path = ""
        project_config_path = ""

        if self._has_file(self.__user_config_file):
            LOGGER.debug("Found User CFNLINTRC")
            user_config_path = self.__user_config_file

        if self._has_file(self.__project_config_file):
            LOGGER.debug("Found Project level CFNLINTRC")
            project_config_path = self.__project_config_file

        return user_config_path, project_config_path

    def _has_file(self, filename):
        """Confirm whether file exists
        Parameters
        ----------
        filename : str
            Path to a file
        Returns
        -------
        Boolean
        """

        return Path(filename).is_file()

    def load(self):
        """Load configuration file and expose as a dictionary
        Returns
        -------
        Dict
            CFLINTRC configuration
        """

        if self.__custom_config_file:
            custom_config = self._read_config(self.__custom_config_file)
            LOGGER.debug("Validating Custom CFNLINTRC")
            self.validate_config(custom_config, self.schema)
            LOGGER.debug("Custom configuration loaded as")
            LOGGER.debug("%s", custom_config)

            self.file_args = custom_config
        else:
            user_config = self._read_config(self.__user_config_file)
            LOGGER.debug("Validating User CFNLINTRC")
            self.validate_config(user_config, self.schema)

            project_config = self._read_config(self.__project_config_file)
            LOGGER.debug("Validating Project CFNLINTRC")
            self.validate_config(project_config, self.schema)

            LOGGER.debug("User configuration loaded as")
            LOGGER.debug("%s", user_config)
            LOGGER.debug("Project configuration loaded as")
            LOGGER.debug("%s", project_config)

            LOGGER.debug("Merging configurations...")
            self.file_args = self.merge_config(user_config, project_config)

    def validate_config(self, config, schema):
        """Validate configuration against schema
        Parameters
        ----------
        config : dict
            CFNLINTRC configuration
        schema : dict
            JSONSchema to validate against
        Raises
        -------
        jsonschema.exceptions.ValidationError
            Returned when cfnlintrc doesn't match schema provided
        """
        LOGGER.debug("Validating CFNLINTRC config with given JSONSchema")
        LOGGER.debug("Schema used: %s", schema)
        LOGGER.debug("Config used: %s", config)

        jsonschema.validate(config, schema)
        LOGGER.debug("CFNLINTRC looks valid!")

    def merge_config(self, user_config, project_config):
        """Merge project and user configuration into a single dictionary
        Creates a new configuration with both configuration merged
        it favours project level over user configuration if keys are duplicated
        NOTE
        ----
            It takes any number of nested dicts
            It overrides lists found in user_config with project_config
        Parameters
        ----------
        user_config : Dict
            User configuration (~/.cfnlintrc) found at user's home directory
        project_config : Dict
            Project configuration (.cfnlintrc) found at current directory
        Returns
        -------
        Dict
            Merged configuration
        """
        # Recursively override User config with Project config
        for key in user_config:
            if key in project_config:
                # If both keys are the same, let's check whether they have nested keys
                if isinstance(user_config[key], dict) and isinstance(
                    project_config[key], dict
                ):
                    self.merge_config(user_config[key], project_config[key])
                else:
                    user_config[key] = project_config[key]
                    LOGGER.debug(
                        "Overriding User's key %s with Project's specific value %s.",
                        key,
                        project_config[key],
                    )

        # Project may have unique config we need to copy over too
        # so that we can have user+project config available as one
        for key in project_config:
            if key not in user_config:
                user_config[key] = project_config[key]

        return user_config

    def _read_config(self, config):
        """Parse given YAML configuration
        Returns
        -------
        Dict
            Parsed YAML configuration as dictionary
        """
        config = Path(config)
        config_template = None

        if self._has_file(config):
            LOGGER.debug("Parsing CFNLINTRC")
            config_template = cfnlint.decode.cfn_yaml.load(str(config))

        if not config_template:
            config_template = {}

        return config_template


def comma_separated_arg(string):
    """Split a comma separated string"""
    return string.split(",")


def _ensure_value(namespace, name, value):
    if getattr(namespace, name, None) is None:
        setattr(namespace, name, value)
    return getattr(namespace, name)


class RuleConfigurationAction(argparse.Action):
    """Override the default Action"""

    def __init__(
        self,
        option_strings,
        dest,
        nargs=None,
        const=None,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):  # pylint: disable=W0622
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def _parse_rule_configuration(self, string):
        """Parse the config rule structure"""
        configs = comma_separated_arg(string)
        results = {}
        for config in configs:
            rule_id = config.split(":")[0]
            config_name = config.split(":")[1].split("=")[0]
            config_value = config.split(":")[1].split("=")[1]
            if rule_id not in results:
                results[rule_id] = {}
            results[rule_id][config_name] = config_value

        return results

    def __call__(self, parser, namespace, values, option_string=None):
        items = copy.copy(_ensure_value(namespace, self.dest, {}))
        try:
            for value in values:
                new_value = self._parse_rule_configuration(value)
                for v_k, v_vs in new_value.items():
                    if v_k in items:
                        for s_k, s_v in v_vs.items():
                            items[v_k][s_k] = s_v
                    else:
                        items[v_k] = v_vs
            setattr(namespace, self.dest, items)
        except Exception:  # pylint: disable=W0703
            parser.print_help()
            parser.exit()


class CliArgs:
    """Base Args class"""

    cli_args: Dict = {}

    def __init__(self, cli_args):
        self.parser = self.create_parser()
        self.cli_args, _ = self.parser.parse_known_args(cli_args)

    def create_parser(self):
        """Do first round of parsing parameters to set options"""

        class ArgumentParser(argparse.ArgumentParser):
            """Override Argument Parser so we can control the exit code"""

            def error(self, message):
                self.print_help(sys.stderr)
                self.exit(32, f"{self.prog}: error: {message}\n")

        class ExtendAction(argparse.Action):
            """Support argument types that are lists and can be specified multiple times."""

            def __call__(self, parser, namespace, values, option_string=None):
                items = getattr(namespace, self.dest)
                items = [] if items is None else items
                for value in values:
                    if isinstance(value, list):
                        items.extend(value)
                    else:
                        items.append(value)
                setattr(namespace, self.dest, items)

        usage = (
            "\nBasic: cfn-lint test.yaml\n"
            "Ignore a rule: cfn-lint -i E3012 -- test.yaml\n"
            "Configure a rule: cfn-lint -x E3012:strict=false -t test.yaml\n"
            "Lint all yaml files in a folder: cfn-lint dir/**/*.yaml"
        )

        parser = ArgumentParser(description="CloudFormation Linter", usage=usage)
        parser.register("action", "extend", ExtendAction)

        standard = parser.add_argument_group("Standard")
        advanced = parser.add_argument_group("Advanced / Debugging")

        # Allow the template to be passes as an optional or a positional argument
        standard.add_argument(
            "templates",
            metavar="TEMPLATE",
            nargs="*",
            help="The CloudFormation template to be linted",
        )
        standard.add_argument(
            "-t",
            "--template",
            metavar="TEMPLATE",
            dest="template_alt",
            help="The CloudFormation template to be linted",
            nargs="+",
            default=[],
            action="extend",
        )
        standard.add_argument(
            "-b",
            "--ignore-bad-template",
            help="Ignore failures with Bad template",
            action="store_true",
        )
        standard.add_argument(
            "--ignore-templates",
            dest="ignore_templates",
            help="Ignore templates",
            nargs="+",
            default=[],
            action="extend",
        )
        advanced.add_argument(
            "-D", "--debug", help="Enable debug logging", action="store_true"
        )
        advanced.add_argument(
            "-I", "--info", help="Enable information logging", action="store_true"
        )
        standard.add_argument(
            "-f",
            "--format",
            help="Output Format",
            choices=["quiet", "parseable", "json", "junit", "pretty", "sarif"],
        )
        standard.add_argument(
            "-l",
            "--list-rules",
            dest="listrules",
            default=False,
            action="store_true",
            help="list all the rules",
        )
        standard.add_argument(
            "-r",
            "--regions",
            dest="regions",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="list the regions to validate against.",
        )
        advanced.add_argument(
            "-a",
            "--append-rules",
            dest="append_rules",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="specify one or more rules directories using "
            "one or more --append-rules arguments. ",
        )
        standard.add_argument(
            "-i",
            "--ignore-checks",
            dest="ignore_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="only check rules whose id do not match these values",
        )
        standard.add_argument(
            "-c",
            "--include-checks",
            dest="include_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="include rules whose id match these values",
        )
        standard.add_argument(
            "-m",
            "--mandatory-checks",
            dest="mandatory_checks",
            nargs="+",
            default=[],
            type=comma_separated_arg,
            action="extend",
            help="always check rules whose id match these values, regardless of template exclusions",
        )
        standard.add_argument(
            "-e",
            "--include-experimental",
            help="Include experimental rules",
            action="store_true",
        )
        standard.add_argument(
            "-x",
            "--configure-rule",
            dest="configure_rules",
            nargs="+",
            default={},
            action=RuleConfigurationAction,
            help="Provide configuration for a rule. Format RuleId:key=value. Example: E3012:strict=false",
        )
        standard.add_argument(
            "--config-file",
            dest="config_file",
            help="Specify the cfnlintrc file to use",
        )
        standard.add_argument(
            "-z",
            "--custom-rules",
            dest="custom_rules",
            help="Allows specification of a custom rule file.",
        )
        advanced.add_argument(
            "-o",
            "--override-spec",
            dest="override_spec",
            help="A CloudFormation Spec override file that allows customization",
        )
        advanced.add_argument(
            "-g",
            "--build-graph",
            help="Creates a file in the same directory as the template that models the template's resources in DOT format",
            action="store_true",
        )
        advanced.add_argument(
            "-s",
            "--registry-schemas",
            help="one or more directories of CloudFormation Registry Schemas",
            action="extend",
            type=comma_separated_arg,
            nargs="+",
        )
        standard.add_argument(
            "-v",
            "--version",
            help="Version of cfn-lint",
            action="version",
            version=f"%(prog)s {__version__}",
        )
        advanced.add_argument(
            "-u",
            "--update-specs",
            help="Update the CloudFormation Specs",
            action="store_true",
        )
        advanced.add_argument(
            "--update-documentation", help=argparse.SUPPRESS, action="store_true"
        )
        advanced.add_argument(
            "--update-iam-policies", help=argparse.SUPPRESS, action="store_true"
        )
        standard.add_argument(
            "--output-file",
            type=str,
            default=None,
            help="Writes the output to the specified file, ideal for producing reports",
        )
        standard.add_argument(
            "--merge-configs",
            default=False,
            action="store_true",
            help="Merges lists between configuration layers",
        )
        standard.add_argument(
            "--non-zero-exit-code",
            type=str,
            default="informational",
            choices=["informational", "warning", "error", "none"],
            help="Exit code will be non zero from the specified rule class and higher",
        )
        advanced.add_argument("--force", help=argparse.SUPPRESS, action="store_true")

        return parser


class TemplateArgs:
    """Per Template Args"""

    def __init__(self, template_args):
        self.set_template_args(template_args)

    def get_template_args(self):
        return self._template_args

    def set_template_args(self, template):
        defaults = {}
        if isinstance(template, dict):
            metadata = template.get("Metadata", {})
            if metadata:
                configs = (
                    template.get("Metadata", {}).get("cfn-lint", {}).get("config", {})
                )

                if isinstance(configs, dict):
                    for config_name, config_value in configs.items():
                        if config_name == "ignore_checks":
                            if isinstance(config_value, list):
                                defaults["ignore_checks"] = config_value
                        if config_name == "regions":
                            if isinstance(config_value, list):
                                defaults["regions"] = config_value
                        if config_name == "append_rules":
                            if isinstance(config_value, list):
                                defaults["append_rules"] = config_value
                        if config_name == "override_spec":
                            if isinstance(config_value, (str)):
                                defaults["override_spec"] = config_value
                        if config_name == "custom_rules":
                            if isinstance(config_value, (str)):
                                defaults["custom_rules"] = config_value
                        if config_name == "ignore_bad_template":
                            if isinstance(config_value, bool):
                                defaults["ignore_bad_template"] = config_value
                        if config_name == "include_checks":
                            if isinstance(config_value, list):
                                defaults["include_checks"] = config_value
                        if config_name == "configure_rules":
                            if isinstance(config_value, dict):
                                defaults["configure_rules"] = config_value

        self._template_args = defaults

    template_args = property(get_template_args, set_template_args)


# pylint: disable=too-many-public-methods
class ConfigMixIn(TemplateArgs, CliArgs, ConfigFileArgs):
    """Mixin for the Configs"""

    def __init__(self, cli_args):
        CliArgs.__init__(self, cli_args)
        # configure debug as soon as we can
        configure_logging(self.cli_args.debug, self.cli_args.info)
        TemplateArgs.__init__(self, {})
        ConfigFileArgs.__init__(
            self, config_file=self._get_argument_value("config_file", False, False)
        )

    def _get_argument_value(self, arg_name, is_template, is_config_file):
        cli_value = getattr(self.cli_args, arg_name)
        template_value = self.template_args.get(arg_name)
        file_value = self.file_args.get(arg_name)

        # merge list configurations
        # make sure we don't do an infinite loop so skip this check for merge_configs
        if arg_name != "merge_configs":
            if self.merge_configs:
                # the CLI will always have an empty list when the item is a list
                # we will use that to evaluate if we need to merge the lists
                if isinstance(cli_value, list):
                    # Use a copy here, otherwise we will accumulate template level config
                    # into the cli_value which will  persist between template files
                    result = cli_value.copy()
                    if isinstance(template_value, list):
                        result.extend(template_value)
                    if isinstance(file_value, list):
                        result.extend(file_value)
                    return result

        # return individual items
        if cli_value:
            return cli_value
        if template_value and is_template:
            return template_value
        if file_value and is_config_file:
            return file_value
        return cli_value

    @property
    def ignore_checks(self):
        return self._get_argument_value("ignore_checks", True, True)

    @property
    def include_checks(self):
        results = self._get_argument_value("include_checks", True, True)
        return ["W", "E"] + results

    @property
    def mandatory_checks(self):
        return self._get_argument_value("mandatory_checks", False, True)

    @property
    def include_experimental(self):
        return self._get_argument_value("include_experimental", True, True)

    @property
    def regions(self):
        results = self._get_argument_value("regions", True, True)
        if not results:
            default_region_env = os.environ.get("AWS_DEFAULT_REGION", "us-east-1")
            return [os.environ.get("AWS_REGION", default_region_env)]
        if "ALL_REGIONS" in results:
            return REGIONS
        return results

    @property
    def ignore_bad_template(self):
        return self._get_argument_value("ignore_bad_template", True, True)

    @property
    def debug(self):
        return self._get_argument_value("debug", False, False)

    @property
    def format(self):
        return self._get_argument_value("format", False, True)

    @property
    def templates(self):
        templates_args = self._get_argument_value("templates", False, True)
        template_alt_args = self._get_argument_value("template_alt", False, False)
        if template_alt_args:
            filenames = template_alt_args
        elif templates_args:
            filenames = templates_args
        else:
            return None

        # if only one is specified convert it to array
        if isinstance(filenames, str):
            filenames = [filenames]

        # handle different shells and Config files
        # some shells don't expand * and configparser won't expand wildcards
        all_filenames = []
        ignore_templates = self._ignore_templates()
        for filename in filenames:
            add_filenames = glob.glob(filename, recursive=True)
            # only way to know of the glob failed is to test it
            # then add the filename as requested
            if not add_filenames:
                if filename not in ignore_templates:
                    all_filenames.append(filename)
            else:
                for add_filename in add_filenames:
                    if add_filename not in ignore_templates:
                        all_filenames.append(add_filename)

        return sorted(all_filenames)

    def _ignore_templates(self):
        ignore_template_args = self._get_argument_value("ignore_templates", False, True)
        if ignore_template_args:
            filenames = ignore_template_args
        else:
            return []

        # if only one is specified convert it to array
        if isinstance(filenames, str):
            filenames = [filenames]

        # handle different shells and Config files
        # some shells don't expand * and configparser won't expand wildcards
        all_filenames = []
        for filename in filenames:
            add_filenames = glob.glob(filename, recursive=True)
            # only way to know of the glob failed is to test it
            # then add the filename as requested
            if not add_filenames:
                all_filenames.append(filename)
            else:
                all_filenames.extend(add_filenames)

        return all_filenames

    @property
    def append_rules(self):
        return self._get_argument_value("append_rules", False, True)

    @property
    def override_spec(self):
        return self._get_argument_value("override_spec", False, True)

    @property
    def custom_rules(self):
        """custom_rules_spec"""
        return self._get_argument_value("custom_rules", False, True)

    @property
    def update_specs(self):
        return self._get_argument_value("update_specs", False, False)

    @property
    def update_documentation(self):
        return self._get_argument_value("update_documentation", False, False)

    @property
    def update_iam_policies(self):
        return self._get_argument_value("update_iam_policies", False, False)

    @property
    def listrules(self):
        return self._get_argument_value("listrules", False, False)

    @property
    def configure_rules(self):
        return self._get_argument_value("configure_rules", True, True)

    @property
    def config_file(self):
        return self._get_argument_value("config_file", False, False)

    @property
    def build_graph(self):
        return self._get_argument_value("build_graph", False, False)

    @property
    def output_file(self):
        return self._get_argument_value("output_file", False, True)

    @property
    def registry_schemas(self):
        return self._get_argument_value("registry_schemas", False, True)

    @property
    def merge_configs(self):
        return self._get_argument_value("merge_configs", True, True)

    @property
    def non_zero_exit_code(self):
        return self._get_argument_value("non_zero_exit_code", False, False)

    @property
    def force(self):
        return self._get_argument_value("force", False, False)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json
from datetime import date

from cfnlint.helpers import FUNCTIONS_SINGLE, convert_dict
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Policy(CloudFormationLintRule):
    """Check if IAM Policy JSON is correct"""

    id = "E2507"
    shortdesc = "Check if IAM Policies are properly configured"
    description = "See if there elements inside an IAM policy are correct"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-iam-policy.html"
    tags = ["properties", "iam"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_exceptions = {
            "AWS::ECR::Repository": "RepositoryPolicyText",
        }
        self.resources_and_keys = {
            "AWS::ECR::Repository": "RepositoryPolicyText",
            "AWS::Elasticsearch::Domain": "AccessPolicies",
            "AWS::OpenSearchService::Domain": "AccessPolicies",
            "AWS::KMS::Key": "KeyPolicy",
            "AWS::S3::BucketPolicy": "PolicyDocument",
            "AWS::SNS::TopicPolicy": "PolicyDocument",
            "AWS::SQS::QueuePolicy": "PolicyDocument",
        }
        self.idp_and_keys = {
            "AWS::IAM::Group": "Policies",
            "AWS::IAM::ManagedPolicy": "PolicyDocument",
            "AWS::IAM::Policy": "PolicyDocument",
            "AWS::IAM::Role": "Policies",
            "AWS::IAM::User": "Policies",
            "AWS::SSO::PermissionSet": "InlinePolicy",
        }
        for resource_type in self.resources_and_keys:
            self.resource_property_types.append(resource_type)
        for resource_type in self.idp_and_keys:
            self.resource_property_types.append(resource_type)

    def check_policy_document(
        self, value, path, is_identity_policy, resource_exceptions, start_mark, end_mark
    ):
        """Check policy document"""
        matches = []

        valid_keys = [
            "Version",
            "Id",
            "Statement",
        ]
        valid_versions = [
            "2012-10-17",
            "2008-10-17",
            date(2012, 10, 17),
            date(2008, 10, 17),
        ]

        if isinstance(value, str):
            try:
                value = convert_dict(json.loads(value), start_mark, end_mark)
            except Exception as ex:  # pylint: disable=W0703,W0612
                message = "IAM Policy Documents need to be JSON"
                matches.append(RuleMatch(path[:], message))
                return matches

        if not isinstance(value, dict):
            message = "IAM Policy Documents needs to be JSON"
            matches.append(RuleMatch(path[:], message))
            return matches

        for p_vs, p_p in value.items_safe(path[:], (dict)):
            for parent_key, parent_value in p_vs.items():
                if parent_key not in valid_keys:
                    message = f"IAM Policy key {parent_key} doesn't exist."
                    matches.append(RuleMatch(path[:] + p_p + [parent_key], message))
                if parent_key == "Version":
                    if parent_value not in valid_versions:
                        message = f'IAM Policy Version needs to be one of ({", ".join(map(str, ["2012-10-17", "2008-10-17"]))}).'
                        matches.append(RuleMatch(p_p + [parent_key], message))
                if parent_key == "Statement":
                    if isinstance(parent_value, list):
                        for i_s_v, i_s_p in parent_value.items_safe(
                            p_p + ["Statement"], (dict)
                        ):
                            matches.extend(
                                self._check_policy_statement(
                                    i_s_p,
                                    i_s_v,
                                    is_identity_policy,
                                    resource_exceptions,
                                )
                            )
                    elif isinstance(parent_value, dict):
                        for i_s_v, i_s_p in parent_value.items_safe(
                            p_p + ["Statement"]
                        ):
                            matches.extend(
                                self._check_policy_statement(
                                    i_s_p,
                                    i_s_v,
                                    is_identity_policy,
                                    resource_exceptions,
                                )
                            )
                    else:
                        message = "IAM Policy statement should be of list."
                        matches.append(RuleMatch(p_p + [parent_key], message))
        return matches

    def _check_policy_statement(
        self, branch, statement, is_identity_policy, resource_exceptions
    ):
        """Check statements"""
        matches = []
        statement_valid_keys = [
            "Action",
            "Condition",
            "Effect",
            "NotAction",
            "NotPrincipal",
            "NotResource",
            "Principal",
            "Resource",
            "Sid",
        ]

        for key, _ in statement.items():
            if key not in statement_valid_keys:
                message = f"IAM Policy statement key {key} isn't valid"
                matches.append(RuleMatch(branch[:] + [key], message))
        if "Effect" not in statement:
            message = "IAM Policy statement missing Effect"
            matches.append(RuleMatch(branch[:], message))
        else:
            for effect, effect_path in statement.get_safe("Effect"):
                if isinstance(effect, str):
                    if effect not in ["Allow", "Deny"]:
                        message = "IAM Policy Effect should be Allow or Deny"
                        matches.append(RuleMatch(branch[:] + effect_path, message))
        if "Action" not in statement and "NotAction" not in statement:
            message = "IAM Policy statement missing Action or NotAction"
            matches.append(RuleMatch(branch[:], message))
        if is_identity_policy:
            if "Principal" in statement or "NotPrincipal" in statement:
                message = "IAM Resource Policy statement shouldn't have Principal or NotPrincipal"
                matches.append(RuleMatch(branch[:], message))
        else:
            if "Principal" not in statement and "NotPrincipal" not in statement:
                message = "IAM Resource Policy statement should have Principal or NotPrincipal"
                matches.append(RuleMatch(branch[:] + ["Principal"], message))
        if not resource_exceptions:
            if "Resource" not in statement and "NotResource" not in statement:
                message = "IAM Policy statement missing Resource or NotResource"
                matches.append(RuleMatch(branch[:], message))

        resources = statement.get("Resource", [])
        if isinstance(resources, str):
            resources = [resources]

        for index, resource in enumerate(resources):
            if isinstance(resource, dict):
                if len(resource) == 1:
                    for k in resource.keys():
                        if k not in FUNCTIONS_SINGLE:
                            message = (
                                "IAM Policy statement Resource incorrectly formatted"
                            )
                            matches.append(
                                RuleMatch(branch[:] + ["Resource", index], message)
                            )
                else:
                    message = "IAM Policy statement Resource incorrectly formatted"
                    matches.append(RuleMatch(branch[:] + ["Resource", index], message))

        return matches

    def match_resource_properties(self, properties, resourcetype, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        is_identity_policy = True
        if resourcetype in self.resources_and_keys:
            is_identity_policy = False

        key = None
        if resourcetype in self.resources_and_keys:
            key = self.resources_and_keys.get(resourcetype)
        else:
            key = self.idp_and_keys.get(resourcetype)

        if not key:
            # Key isn't defined return nothing
            return matches

        resource_exceptions = False
        if key == self.resource_exceptions.get(resourcetype):
            resource_exceptions = True

        other_keys = []
        for key, value in self.resources_and_keys.items():
            if value != "Policies":
                other_keys.append(key)
        for key, value in self.idp_and_keys.items():
            if value != "Policies":
                other_keys.append(key)

        for key, value in properties.items():
            if key == "Policies" and isinstance(value, list):
                for index, policy in enumerate(properties.get(key, [])):
                    matches.extend(
                        cfn.check_value(
                            obj=policy,
                            key="PolicyDocument",
                            path=path[:] + ["Policies", index],
                            check_value=self.check_policy_document,
                            is_identity_policy=is_identity_policy,
                            resource_exceptions=resource_exceptions,
                            start_mark=key.start_mark,
                            end_mark=key.end_mark,
                        )
                    )
            elif key in [
                "KeyPolicy",
                "PolicyDocument",
                "RepositoryPolicyText",
                "AccessPolicies",
                "InlinePolicy",
            ]:
                matches.extend(
                    cfn.check_value(
                        obj=properties,
                        key=key,
                        path=path[:],
                        check_value=self.check_policy_document,
                        is_identity_policy=is_identity_policy,
                        resource_exceptions=resource_exceptions,
                        start_mark=key.start_mark,
                        end_mark=key.end_mark,
                    )
                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class FargateDeploymentSchedulingStrategy(CloudFormationLintRule):
    id = "E3044"
    shortdesc = "Check Fargate service scheduling strategy"
    description = "Check that Fargate service scheduling strategy is REPLICA"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ecs-service.html#cfn-ecs-service-schedulingstrategy"
    tags = ["properties", "ecs", "service", "container", "fargate"]

    def match(self, cfn):
        matches = []
        ecs_services = cfn.get_resource_properties(["AWS::ECS::Service"])

        for ecs_service in ecs_services:
            path = ecs_service["Path"]
            properties = ecs_service["Value"]
            if isinstance(properties, dict):
                scenarios = cfn.get_object_without_conditions(
                    properties, ["LaunchType", "SchedulingStrategy"]
                )
                for scenario in scenarios:
                    props = scenario.get("Object")
                    launch_type = props.get("LaunchType", None)
                    if isinstance(launch_type, str) and launch_type == "FARGATE":
                        scheduling_strategy = props.get("SchedulingStrategy", None)
                        if (
                            isinstance(scheduling_strategy, str)
                            and scheduling_strategy != "REPLICA"
                        ):
                            error_message = f"Fargate service only support REPLICA as scheduling strategy at {'/'.join(map(str, path))}"
                            matches.append(RuleMatch(path, error_message))
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Configuration(CloudFormationLintRule):
    """Check if Mappings are configured correctly"""

    id = "E7001"
    shortdesc = "Mappings are appropriately configured"
    description = "Check if Mappings are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/mappings-section-structure.html"
    tags = ["mappings"]

    def match(self, cfn):
        matches = []

        valid_map_types = (str, list, int, float)

        mappings = cfn.template.get("Mappings", {})
        if mappings:
            for mapname, mapobj in mappings.items():
                if not isinstance(mapobj, dict):
                    message = "Mapping {0} has invalid property"
                    matches.append(
                        RuleMatch(["Mappings", mapname], message.format(mapname))
                    )
                else:
                    for firstkey in mapobj:
                        firstkeyobj = mapobj[firstkey]
                        if not isinstance(firstkeyobj, dict):
                            message = "Mapping {0} has invalid property at {1}"
                            matches.append(
                                RuleMatch(
                                    ["Mappings", mapname, firstkey],
                                    message.format(mapname, firstkeyobj),
                                )
                            )
                        else:
                            for secondkey in firstkeyobj:
                                if not isinstance(
                                    firstkeyobj[secondkey], valid_map_types
                                ):
                                    message = "Mapping {0} has invalid property at {1}"
                                    matches.append(
                                        RuleMatch(
                                            ["Mappings", mapname, firstkey, secondkey],
                                            message.format(mapname, secondkey),
                                        )
                                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class InterfaceParameterExists(CloudFormationLintRule):
    """Check if Metadata Interface parameters exist"""

    id = "W4001"
    shortdesc = "Metadata Interface parameters exist"
    description = "Metadata Interface parameters actually exist"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-cloudformation-interface.html"
    tags = ["metadata"]

    valid_keys = ["ParameterGroups", "ParameterLabels"]

    def match(self, cfn):
        """Check CloudFormation Metadata Parameters Exist"""

        matches = []

        strinterface = "AWS::CloudFormation::Interface"
        parameters = cfn.get_parameter_names()
        metadata_obj = cfn.template.get("Metadata", {})
        if metadata_obj:
            interfaces = metadata_obj.get(strinterface, {})
            if isinstance(interfaces, dict):
                # Check Parameter Group Parameters
                paramgroups = interfaces.get("ParameterGroups", [])
                if isinstance(paramgroups, list):
                    for index, value in enumerate(paramgroups):
                        if "Parameters" in value:
                            for paramindex, paramvalue in enumerate(
                                value["Parameters"]
                            ):
                                if paramvalue not in parameters:
                                    message = (
                                        "Metadata Interface parameter doesn't exist {0}"
                                    )
                                    matches.append(
                                        RuleMatch(
                                            [
                                                "Metadata",
                                                strinterface,
                                                "ParameterGroups",
                                                index,
                                                "Parameters",
                                                paramindex,
                                            ],
                                            message.format(paramvalue),
                                        )
                                    )
                paramlabels = interfaces.get("ParameterLabels", {})
                if isinstance(paramlabels, dict):
                    for param in paramlabels:
                        if param not in parameters:
                            message = "Metadata Interface parameter doesn't exist {0}"
                            matches.append(
                                RuleMatch(
                                    [
                                        "Metadata",
                                        strinterface,
                                        "ParameterLabels",
                                        param,
                                    ],
                                    message.format(param),
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from cfnlint.helpers import PSEUDOPARAMS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RelationshipConditions(CloudFormationLintRule):
    """Check if Ref/GetAtt values are available via conditions"""

    id = "W1001"
    shortdesc = "Ref/GetAtt to resource that is available when conditions are applied"
    description = (
        "Check the Conditions that affect a Ref/GetAtt to make sure "
        "the resource being related to is available when there is a resource "
        "condition."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["conditions", "resources", "relationships", "ref", "getatt", "sub"]

    def match(self, cfn):
        """Check CloudFormation Ref/GetAtt for Conditions"""

        matches = []

        # Start with Ref checks
        ref_objs = cfn.search_deep_keys(searchText="Ref", includeGlobals=False)
        for ref_obj in ref_objs:
            value = ref_obj[-1]
            if value not in PSEUDOPARAMS:
                scenarios = cfn.is_resource_available(ref_obj, value)
                for scenario in scenarios:
                    # pylint: disable=consider-using-f-string
                    scenario_text = " and ".join(
                        [
                            'when condition "%s" is %s' % (k, v)
                            for (k, v) in scenario.items()
                        ]
                    )
                    message = (
                        'Ref to resource "{0}" that may not be available {1} at {2}'
                    )
                    matches.append(
                        RuleMatch(
                            ref_obj[:-1],
                            message.format(
                                value, scenario_text, "/".join(map(str, ref_obj[:-1]))
                            ),
                        )
                    )

        # The do GetAtt
        getatt_objs = cfn.search_deep_keys(
            searchText="Fn::GetAtt", includeGlobals=False
        )
        for getatt_obj in getatt_objs:
            value_obj = getatt_obj[-1]
            value = None
            if isinstance(value_obj, list):
                value = value_obj[0]
            elif isinstance(value_obj, str):
                value = value_obj.split(".")[0]
            if value:
                if value not in PSEUDOPARAMS:
                    scenarios = cfn.is_resource_available(getatt_obj, value)
                    for scenario in scenarios:
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        message = 'GetAtt to resource "{0}" that may not be available {1} at {2}'
                        matches.append(
                            RuleMatch(
                                getatt_obj[:-1],
                                message.format(
                                    value,
                                    scenario_text,
                                    "/".join(map(str, getatt_obj[:-1])),
                                ),
                            )
                        )

        # The do Sub
        sub_objs = cfn.search_deep_keys(searchText="Fn::Sub", includeGlobals=False)
        for sub_obj in sub_objs:
            sub_string = sub_obj[-1]
            # Filter out bad types of sub_strings.
            # Lists have two be two items and it can be just a string
            if not isinstance(sub_string, (list, str)):
                continue
            if isinstance(sub_string, str):
                sub_string = [sub_string, {}]
            if len(sub_string) != 2:
                continue
            sub_params = sub_string[1]
            string_params = cfn.get_sub_parameters(sub_string[0])

            for string_param in string_params:
                if string_param not in sub_params:
                    # deal with GetAtts by dropping everything after the .
                    string_param = string_param.split(".")[0]
                    if string_param in cfn.template.get("Resources", {}):
                        scenarios = cfn.is_resource_available(
                            sub_obj[:-1], string_param
                        )
                        for scenario in scenarios:
                            scenario_text = " and ".join(
                                [
                                    f'when condition "{k}" is {v}'
                                    for (k, v) in scenario.items()
                                ]
                            )
                            message = 'Fn::Sub to resource "{0}" that may not be available {1} at {2}'
                            matches.append(
                                RuleMatch(
                                    sub_obj[:-1],
                                    message.format(
                                        string_param,
                                        scenario_text,
                                        "/".join(map(str, sub_obj[:-1])),
                                    ),
                                )
                            )

        return matches
